# -*- coding: utf-8 -*-

import os
import sys
import threading
try:
    import Queue
except ImportError:
    import queue as Queue
import time
import tempfile
import string
import logging
import itertools
import re

import six

from .. import runners
from .. import picklerunner

from ..helpers import format_command
from ..helpers import file_size

if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess

class GridJobRequires(object):
    """Defines the resources required for a task on the grid.

    :param time: Wall clock time in minutes.
    :type time: int

    :param mem: RAM Usage in MB (8*1024*1024 bits).
    :type mem: int

    :param cores: CPU cores.
    :type cores: int
    
    :param partition: grid partition.
    :type partition: string
    """
    
    def __init__(self, time, mem, cores, partition, docker_image, depends=None):
        # if time is not an int, try to format the equation
        if not str(time).isdigit():
            self.time = format_command(time, depends=depends, cores=cores)
        else:
            self.time = int(time)
        
        # if memory is not an int, try to format the equation
        if not str(mem).isdigit():
            self.mem = format_command(mem, depends=depends, cores=cores) 
        else:
            self.mem = int(mem)
            
        self.cores = int(cores) 
        
        self.partition = partition
        self.docker_image = docker_image
    
class Grid(object):
    """ Base Grid Workflow manager class """
    
    def __init__(self, name, worker, queue, tmpdir, benchmark_on=None):
        self.name = name
        self.worker = worker
        self.queue = queue
        self.tmpdir = tmpdir
        
        # create the folder if it does not already exist for temp directory
        if not os.path.isdir(self.tmpdir):
            os.makedirs(self.tmpdir)
        
        self.task_data = dict()

    def _get_grid_task_settings(self, kwargs, depends):
        """ Get the resources required to run this task on the grid """
        # check for the required keywords
        requires=[]
        for key in ["time","mem","cores"]:
            try:
                requires.append(kwargs[key])
            except KeyError:
                raise KeyError(key+" is a required keyword argument for a grid task")

        # check for optional keyword
        try:
            requires.append(kwargs["partition"])
        except KeyError:
            requires.append(None)    
        
        try:
            requires.append(kwargs["docker_image"])
        except KeyError:
            requires.append(None)    
        
        requires+=[depends]
        
        return (GridJobRequires(*requires), self.tmpdir)
        
    def do(self, task, **kwargs):
        """Accepts the following extra arguments:
        
        :param time: The maximum time in minutes allotted to run the
          command
        :type time: int

        :param mem: The maximum memory in megabytes allocated to run
          the command
        :type mem: int

        :param cores: The number of CPU cores allocated to the job
        :type cores: int

        :param partition: The grid partition to send this job to
        :type partition: str
        """
        
        self.add_task(task, **kwargs)


    def add_task(self, task, **kwargs):
        """Accepts the following extra arguments:
        
        :keyword time: The maximum time in minutes allotted to run the
          command
        :type time: int

        :keyword mem: The maximum memory in megabytes allocated to run
          the command
        :type mem: int

        :keyword cores: The number of CPU cores allocated to the job
        :type cores: int

        :keyword partition: The grid partition to send this job to
        :type partition: str
        """
        
        self.task_data[task.task_no] = self._get_grid_task_settings(kwargs, task.depends)


    def runner(self, workflow, jobs=1, grid_jobs=1):
        runner = runners.GridRunner(workflow)
        runner.add_worker(runners.ParallelLocalWorker,
                          name="local", rate=jobs, default=True)
        runner.add_worker(self.worker, name=self.name, rate=grid_jobs)
        runner.routes.update((
            ( task_no, (self.name, list(extra)+[self.queue, workflow._reporter]) )
            for task_no, extra in six.iteritems(self.task_data)
        ))
        return runner   

class GridQueue(object):
    
    def __init__(self, partition, benchmark_on=None):
        # check for short/long partitions
        if not isinstance(partition, list):
            partition = [x.strip() for x in partition.split(",")] 
        try:
            self.partition_short, self.partition_long, self.partition_cutoff = partition
            self.partition_cutoff = int(self.partition_cutoff)
        except ValueError:
            self.partition_short = partition[0]
            self.partition_long = partition[0]
            self.partition_cutoff = 0
            
        # this is the refresh rate for checking the queue, in seconds
        self.refresh_rate = 10*60

        # this is the rate for checking the job status, in seconds
        self.check_job_rate = 60
        
        # this is the number of minutes to wait if there is an time out
        # socket error returned from the scheduler when running a command
        self.timeout_sleep = 5*60
        
        # this is the number of times to retry after a timeout error
        self.timeout_retry_max = 3
        
        # this is the number of seconds to wait after job submission
        self.submit_sleep = 5
        
        # this is the last time the queue was checked
        self.last_check = time.time()
        self.sacct = None
        
        # create a lock for jobs in queue
        self.lock_status = threading.Lock()
        self.lock_submit = threading.Lock()
        
        # set if benchmarking should be run
        self.benchmark_on = benchmark_on
        
    @staticmethod
    def submit_command(grid_script):
        raise NotImplementedError
    
    def submit_template(self):
        raise NotImplementedError
    
    def job_failed(self,status):
        raise NotImplementedError
    
    def job_stopped(self,status):
        raise NotImplementedError
    
    def refresh_queue_status(self):
        raise NotImplementedError
    
    def job_memkill(self, status, jobid, memory):
        return False
        
    def job_timeout(self, status, jobid, time):
        return False
    
    def get_job_status_from_stderr(self, error_file, grid_job_status, grid_jobid):
        return grid_job_status
    
    def get_partition(self, time, partition):
        """ Get the partition for the task based on the time requested """
        
        # if a partition is already set for the task, use that partition
        if not partition is None:
            return partition
        
        if time > self.partition_cutoff:
            return self.partition_long
        else:
            return self.partition_short
    
    def get_queue_status(self, refresh=None):
        """ Get the queue accounting stats """
        
        # lock to prevent race conditions with status update
        self.lock_status.acquire()
        
        # check the last time the queue was captured and refresh if set
        current_time = time.time()
        if ( current_time - self.last_check > self.refresh_rate ) or refresh or self.sacct is None:
            self.last_check = current_time
            logging.info("Getting latest queue info to refresh job status")
            self.sacct = self.refresh_queue_status()
            
        self.lock_status.release()
        
        return self.sacct
    
    def get_all_stats_for_jobid(self,jobid):
        """ Get all the stats for a specific job id """
        
        # use the existing stats, to get the information for the jobid
        try:
            job_stats=list(filter(lambda x: x[0].startswith(jobid),self.get_queue_status()))
        except IndexError:
            job_stats=[]
       
        # if the job stats are not found for the job, return an NA state
        if not job_stats:
            job_stats=[[jobid,"Pending","NA","NA","NA"]] 

        return job_stats
    
    def get_job_status(self, jobid):
        """ Check the status of the job """
        
        info=self.get_all_stats_for_jobid(jobid)
        
        return info[0][1]

    def record_benchmark(self, jobid, task_number, reporter):
        """ Check the benchmarking stats of the grid id """
        
        # check if benchmarking is set to off
        if not self.benchmark_on:
            logging.info("Benchmarking is set to off")
            return
        
        reporter.task_grid_status(task_number,jobid,"Getting benchmarking data")
        status, cpus, elapsed, memory = self.get_benchmark(jobid)
            
        logging.info("Benchmark information for job id %s:\nElapsed Time: %s \nCores: %s\nMemory: %s MB",
            task_number, elapsed, cpus, memory)   
        
        reporter.task_grid_status(task_number,jobid,"Final status of "+status)
        
    def get_benchmark(self, jobid, wait=None):
        """ Get the benchmarking data for the jobid """
        
        # if the job is not shown to have finished running then
        # wait for the next queue refresh
        status=self.get_job_status(jobid)
        if wait or not (self.job_stopped(status) or self.job_failed(status)):
            wait_time = abs(self.refresh_rate - (time.time() - self.last_check)) + 10
            time.sleep(wait_time)

        info=self.get_all_stats_for_jobid(jobid)
        
        try:
            status=info[0][1]
        except IndexError:
            status="Unknown"

        try:
            cpus=info[0][2]
        except IndexError:
            cpus="NA"
    
        try:
            elapsed=info[0][3]
        except IndexError:
            elapsed="NA"

        # get the memory max from the batch line which is the second line of output
        try:
            memory=info[0][4]
        except IndexError:
            memory="NA"
        
        if "K" in memory:    
            # if memory is in KB, convert to MB
            memory="{:.1f}".format(float(memory.replace("K",""))/1024.0)
        elif "M" in memory:
            memory="{:.1f}".format(float(memory.replace("M","")))
        elif "G" in memory:
            # if memory is in GB, convert to MB
            memory="{:.1f}".format(float(memory.replace("G",""))*1024.0)
            
        return status, cpus, elapsed, memory
    
    def run_grid_command(self,command):
        """ Run the grid command and check for errors """
       
        error=None
        if six.callable(command):
            try:
                logging.debug("Running grid submit command")
                stdout=command()
            except StandardError as err:
                error=err.output
                stdout=error or "error"
        else: 
            try:
                logging.debug("Running grid command: %s"," ".join(command))
                stdout=subprocess.check_output(command, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as err:
                error=err.output
                stdout=error or "error"
            
        timeout_error=False
        if error and "error" in error and "Socket timed out on send/recv operation" in error:
            # check for a socket timeout error
            timeout_error=True
            
        return stdout, timeout_error
    
    def run_grid_command_resubmit(self,command):
        """ Run this grid command, check for error, resubmit if needed """
        
        # run the grid command
        stdout, timeout_error = self.run_grid_command(command)
        
        # retry if timeout error present after wait
        resubmissions = 0
        if timeout_error and resubmissions < self.timeout_retry_max:
            resubmissions+=1
            # wait before retrying
            logging.warning("Unable to run grid command, waiting and retrying")
            time.sleep(self.timeout_sleep)
            stdout, timeout_error = self.run_grid_command(command)
        
        return stdout
    
    @staticmethod
    def job_submission_failed(jobid):
        """ Check if the job failed in submission and did not get an id """
        return True if not jobid.isdigit() else False

    @staticmethod
    def get_job_id_from_submit_output(stdout):
        try:
            # search for the decimal job id at any location in stdout
            jobid=re.findall(r'\d+',stdout)[0]
        except IndexError:
            jobid="error"

        return jobib

    def submit_job(self,grid_script):
        """ Submit the grid jobs and return the grid job id """
        
        # lock so only one task submits jobs to the queue at a time
        self.lock_submit.acquire()

        # submit the job and get the grid id
        logging.debug("Submitting job to grid")
        stdout=self.run_grid_command_resubmit(self.submit_command(grid_script))

        # get the job id from the stdout
        jobid=self.get_job_id_from_submit_output(stdout)        
        
        # check the jobid for a submission failed
        if self.job_submission_failed(jobid):
            logging.error("Unable to submit job to queue: "+stdout)
        
        # pause for the scheduler
        time.sleep(self.submit_sleep)
        
        self.lock_submit.release()
    
        return jobid
    
    def create_grid_script(self,partition,cpus,minutes,memory,command,taskid,dir,docker_image):
        """ Create a grid script from the template also creating temp stdout and stderr files """
    
        # create temp files for stdout, stderr, and return code    
        handle_out, out_file=tempfile.mkstemp(suffix=".out",prefix="task_"+str(taskid)+"_",dir=dir)
        os.close(handle_out)
        handle_err, error_file=tempfile.mkstemp(suffix=".err",prefix="task_"+str(taskid)+"_",dir=dir)
        os.close(handle_err)
        handle_rc, rc_file=tempfile.mkstemp(suffix=".rc",prefix="task_"+str(taskid)+"_",dir=dir)
        os.close(handle_rc)
        
        # add the remaining sections to the bash template
        bash_template = string.Template("\n".join(["#!/bin/bash "] + self.submit_template() + ["", "${command}", "${rc_command}"]))
    
        # convert the minutes to the time string "HH:MM:00"
        hours, remaining_minutes = divmod(minutes, 60)
        time = "{:02d}:{:02d}:00".format(hours, remaining_minutes)
    
        bash=bash_template.substitute(partition=partition,cpus=cpus,time=time,
            memory=memory,command=command,output=out_file,error=error_file,rc_command="export RC=$? ; echo $RC > "+rc_file+" ; bash -c 'exit $RC'")
        file_handle, new_file=tempfile.mkstemp(suffix=".bash",prefix="task_"+str(taskid)+"_",dir=dir)
        os.write(file_handle,bash)
        os.close(file_handle)
        
        return new_file, out_file, error_file, rc_file


class GridWorker(threading.Thread):
    """ Base Grid Worker class """
    
    def __init__(self, work_q, result_q, lock, reporter):
        super(GridWorker, self).__init__()
        self.daemon = True
        self.logger = runners.logger
        self.work_q = work_q
        self.result_q = result_q
        self.lock = lock
        self.reporter = reporter    
    
    @staticmethod
    def appropriate_q_class(*args, **kwargs):
        return six.moves.queue.Queue(*args, **kwargs)    

    @staticmethod
    def appropriate_lock():
        return threading.Lock() 
    
    def run(self):
        return runners.worker_run_loop(self.work_q, self.result_q, self.run_task_by_type,
            self.reporter, self.lock)
   
    @classmethod 
    def run_task_by_type(cls, task, extra):
        # if any of the tasks are a function, then use pickle interface
        if list(filter(six.callable,task.actions)):
            return cls.run_task_function(task, extra)
        else:
            return cls.run_task_command(task, extra)   
        
    @classmethod
    def run_task_function(cls, task, extra):
        (perf, tmpdir, grid_queue, reporter) = extra
        
        # create a script to run the python function
        pickle_script = picklerunner.PickleScript(task, tmpdir, "task_"+str(task.task_no))
        pickle_task = pickle_script.create_task()
        
        # run the task as a command
        result = cls.run_task_command(pickle_task, extra)
        
        # decode the result
        result = pickle_script.result(result)

        return result
        
    @classmethod
    def run_task_command(cls, task, extra):
        (perf, tmpdir, grid_queue, reporter) = extra
        # report the task has started
        reporter.task_running(task.task_no)
        
        # create a script and stdout/stderr files for this task
        commands="\n".join(task.actions)
        logging.info("Running commands for task id %s:\n%s", task.task_no, commands)
    
        resubmission = 0    
        cores, time, memory, partition, docker_image = perf.cores, perf.time, perf.mem, perf.partition, perf.docker_image
    
        jobid, out_file, error_file, rc_file = cls.submit_grid_job(cores, time, memory, 
            partition, tmpdir, commands, task, grid_queue, reporter, docker_image)
    
        # monitor job if submission was successful
        result, job_final_status = cls.check_submission_then_monitor_grid_job(grid_queue, 
            task, jobid, out_file, error_file, rc_file, reporter)
    
        # if a timeout or memory max, resubmit at most three times
        while ( grid_queue.job_timeout(job_final_status, jobid, time) or grid_queue.job_memkill(job_final_status, jobid, memory) ) and resubmission < 3:
            resubmission+=1
            # increase the memory or the time
            if grid_queue.job_timeout(job_final_status, jobid, time):
                time = "({})*2".format(time) if isinstance(time,str) else time*2
                logging.info("Resubmission number %s of grid job for task id %s with 2x more time: %s minutes", 
                    resubmission, task.task_no, time)
                reporter.task_grid_status(task.task_no,jobid,"Resubmitting due to time out")
            elif grid_queue.job_memkill(job_final_status, jobid, memory):
                memory = "({})*2".format(memory) if isinstance(memory,str) else memory*2
                logging.info("Resubmission number %s of grid job for task id %s with 2x more memory: %s MB",
                    resubmission, task.task_no, memory)
                reporter.task_grid_status(task.task_no,jobid,"Resubmitting due to max memory")
            
            jobid, out_file, error_file, rc_file = cls.submit_grid_job(cores, time, memory,
                partition, tmpdir, commands, task, grid_queue, reporter, docker_image)
    
            # monitor job if submission was successful
            result, job_final_status = cls.check_submission_then_monitor_grid_job(grid_queue, 
                task, jobid, out_file, error_file, rc_file, reporter)
                
        # get the benchmarking data if the job was submitted
        if not grid_queue.job_submission_failed(jobid):
            grid_queue.record_benchmark(jobid, task.task_no, reporter)
        
        return result
    
    @classmethod
    def submit_grid_job(cls, cores, time, memory, partition, tmpdir, commands, task, grid_queue, reporter, docker_image):
        
        # evaluate the time/memory requests for the job
        time, memory = cls.evaluate_resource_requests(time, memory)
        
        # get the partition for the task
        current_partition = grid_queue.get_partition(time, partition)
        
        # create the grid bash script
        grid_script, out_file, error_file, rc_file = grid_queue.create_grid_script(current_partition,
            cores, time, memory, commands, task.task_no, tmpdir, docker_image)
    
        logging.info("Created grid files for task id %s: %s, %s, %s, %s",
            task.task_no, grid_script, out_file, error_file, rc_file)
    
        # submit the job
        jobid = grid_queue.submit_job(grid_script)
        
        logging.info("Submitted job for task id %s: grid id %s", task.task_no,
            jobid)
        
        if not grid_queue.job_submission_failed(jobid):
            reporter.task_grid_status(task.task_no,jobid,"Submitted")
       
        return jobid, out_file, error_file, rc_file
    
    @staticmethod
    def log_grid_output(taskid, file, file_type):
        """ Write the grid stdout/stderr files to the log """
        
        try:
            lines=open(file).readlines()
        except EnvironmentError:
            lines=[]
            
        logging.info("Grid %s from task id %s:\n%s",taskid, file_type, "".join(lines))
        
    @staticmethod
    def get_return_code(file):
        """ Read the return code from the file """
    
        try:
            line=open(file).readline().rstrip()
        except (EnvironmentError, TypeError):
            line=""
    
        return line
    
    @staticmethod
    def evaluate_resource_requests(time,mem):
        """ Evaluate the time/memory requests for the grid job, allowing for ints or formulas """
        
        try:
            time=eval(str(time))
        except TypeError:
            raise TypeError("Unable to evaluate time request for task: "+ time)
        
        try:
            mem=eval(str(mem))
        except TypeError:
            raise TypeError("Unable to evaluate memory request for task: "+ mem)
        
        return time, mem  
    
    @classmethod
    def check_submission_then_monitor_grid_job(cls, grid_queue, task, grid_jobid, 
        out_file, error_file, rc_file, reporter):
        
        # monitor job if submission was successful
        if not grid_queue.job_submission_failed(grid_jobid):
            result, job_final_status = cls.monitor_grid_job(grid_queue, task, grid_jobid,
                out_file, error_file, rc_file, reporter)
        else:
            job_final_status = "SUBMIT FAILED"
            # get the anadama task result
            result=runners._get_task_result(task)
            # add the extra error
            result = result._replace(error=str(result.error)+"Unable to submit job to queue.")
            
        return result, job_final_status
   
    @classmethod 
    def monitor_grid_job(cls, grid_queue, task, grid_jobid, out_file, error_file, rc_file, reporter): 
        # poll to check for status
        grid_job_status=None
        for tries in itertools.count(1):
            # only check status at intervals
            time.sleep(grid_queue.check_job_rate)
            
            # check the queue stats
            grid_job_status = grid_queue.get_job_status(grid_jobid)
            reporter.task_grid_status_polling(task.task_no,grid_jobid,grid_job_status)
            
            logging.info("Status for job id %s with grid id %s is %s",task.task_no,
                grid_jobid,grid_job_status)
            
            if grid_queue.job_stopped(grid_job_status):
                logging.info("Grid status for job id %s shows it has stopped",task.task_no)
                break
            
            # check if the return code file is written
            if rc_file and os.path.getsize(rc_file) > 0:
                logging.info("Return code file for job id %s shows it has stopped",task.task_no)
                break
            
        # check if a grid error is written to the output file
        grid_job_status = grid_queue.get_job_status_from_stderr(error_file, grid_job_status, grid_jobid)
        
        # write the stdout and stderr to the log
        if out_file:
            cls.log_grid_output(task.task_no, out_file, "standard output")
        if error_file:
            cls.log_grid_output(task.task_no, error_file, "standard error")
        if rc_file:
            cls.log_grid_output(task.task_no, rc_file, "return code")
        
        # check the return code
        extra_error=""
        return_code=cls.get_return_code(rc_file)
        if return_code and not return_code == "0":
            extra_error="\nReturn Code Error: " + return_code
          
        # check the queue status
        if grid_queue.job_failed(grid_job_status):
            extra_error+="\nGrid Status Error: " + grid_job_status
     
        # get the anadama task result
        result=runners._get_task_result(task)
    
        # add the extra error if found
        if extra_error:
            result = result._replace(error=str(result.error)+extra_error)
     
        return result, grid_job_status
    
    



# -*- coding: utf-8 -*-
import os
import sys
import json

import six
import leveldb

from .util import mkdirp
from .tracked import s3_folder

ENV_VAR = "ANADAMA_BACKEND_DIR"
LOCAL_DB_FOLDER = ".anadama"

_default_backend = None

def default(output_dir=None):
    global _default_backend
    if s3_folder(output_dir):
        output_dir=os.getcwd()
    if output_dir is not None:
        return LevelDBBackend(
            _try_dir(os.path.abspath(os.path.join(output_dir, LOCAL_DB_FOLDER, "db")))
            )
    if _default_backend is None:
        _default_backend = LevelDBBackend()
    return _default_backend


def discover_data_directory():
    if ENV_VAR in os.environ:
        return _try_dir(os.environ[ENV_VAR])
    elif "HOME" in os.environ:
        return _try_dir(os.path.join(os.environ['HOME'], ".config",
                                     "anadama", "db"))
    else:
        return _fallback_datadir()

def auto(data_dir, *args, **kwargs):
    """Return the right type of backend for the given data_dir"""
    # just one type of backend for now
    return LevelDBBackend(data_dir, *args, **kwargs)


def _try_dir(maybe_datadir):
    if os.path.isdir(maybe_datadir):
        pass
    else:
        try:
            mkdirp(maybe_datadir)
        except Exception as e:
            msg = six.u("Unable to create anadama "
                        "database directory `{}': \n"+str(e))
            sys.stderr.write(msg.format(maybe_datadir))
            fallback = _fallback_datadir()
            sys.stderr.write(six.u("Using fallback directory: "+fallback+"\n"))
            return fallback
    return maybe_datadir

    
def _fallback_datadir():
    try:
        folder=os.path.join(LOCAL_DB_FOLDER,"db")
        mkdirp(folder)
        return os.path.abspath(folder)
    except:
        mkdirp("/tmp/anadama/db")        
        return "/tmp/anadama/db"

    
class BaseBackend(object):
    def __init__(self, data_directory=None, autocreate=True):
        self.data_directory = data_directory or discover_data_directory()
        if autocreate and not self.exists():
            self.create()

    def lookup(self, dep):
        raise NotImplementedError()

    def lookup_many(self, deps):
        raise NotImplementedError()

    def save(self, dep_keys, dep_vals):
        raise NotImplementedError()

    def create(self):
        raise NotImplementedError()

    def exists(self):
        raise NotImplementedError()

    def keys(self):
        raise NotImplementedError()

    def delete(self, key):
        raise NotImplementedError()

    def delete_many(self, keys):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()



class LevelDBBackend(BaseBackend):
    def __init__(self, *args, **kwargs):
        self.db = None
        super(LevelDBBackend, self).__init__(*args, **kwargs)
        if not self.db:
            self.db = leveldb.LevelDB(self.data_directory,
                                      create_if_missing=False)


    def exists(self):
        return all([os.path.exists(os.path.join(self.data_directory, f))
                    for f in ("CURRENT", "LOCK", "LOG")])


    def create(self):
        self.db = leveldb.LevelDB(self.data_directory, create_if_missing=True,
                                  error_if_exists=True)


    def _get(self, key):
        try:
            val = self.db.Get(key.encode("utf-8"))
        except KeyError:
            return None
        return json.loads(val.decode("utf-8"))


    def lookup(self, dep):
        return self._get(dep.name)


    def lookup_many(self, deps):
        return [self.lookup(dep) for dep in deps]


    def save(self, dep_keys, dep_vals):
        if not dep_keys:
            return
        batch = leveldb.WriteBatch()
        for key, val in zip(dep_keys, dep_vals):
            batch.Put(key.encode("utf-8"), json.dumps(val).encode("utf-8"))
        self.db.Write(batch)


    def keys(self):
        return self.db.RangeIter(include_value=False)

    def delete(self, key):
        return self.db.Delete(key.encode("utf-8"))

    def delete_many(self, keys):
        batch = leveldb.WriteBatch()
        for k in keys:
            batch.Delete(k.encode("utf-8"))
        self.db.Write(batch)

    def close(self):
        del self.db
        self.db = None



# -*- coding: utf-8 -*-
import os
import re
import shlex
import fnmatch
import logging
import itertools
from operator import attrgetter, itemgetter
from collections import deque, defaultdict
import copy
import subprocess
import tempfile

import six
from six.moves import filter, map
import networkx as nx
from networkx.algorithms.traversal.depth_first_search import dfs_edges

from . import Task
from . import tracked
from . import grid as _grid
from . import reporters
from . import backends
from . import runners
from .cli import Configuration
from .taskcontainer import TaskContainer
from .helpers import format_command, build_actions
from .util import matcher, noop, find_on_path
from .util import istask, sugar_list, dichotomize
from .util import keepkeys
from .util import fname
from .grid.slurm import Slurm
from .grid.sge import SGE
from .grid.aws import AWS
from .document import PweaveDocument

second = itemgetter(1)
logger = logging.getLogger(__name__)


class RunFailed(ValueError):
    pass


class Workflow(object):
    """Create a Workflow.
    
    :keyword storage_backend: Lookup and save dependency information
      from this object. If ``None`` is passed (the default), the
      default backend from :func:`anadama2.backends.default` is used.
    :type storage_backend: instance of any
      :class:`anadama2.backends.BaseBackend` subclass or None.

    :keyword grid: Use this object to configure the run
      context to submit tasks to a compute grid.  
    :type grid: objects implementing the interface of
      :class:`anadama2.grid.Dummy
    
    :keyword strict: Enable strict mode. If strict, whenever a task is
      added that depends on something that is not the target of
      another task (or isn't marked with
      :math:`anadama2.workflow.Workflow.already_exists`), raise a
      ``KeyError``. If not strict, and the Tracked object
      ``.exists()`` automatically do what's necessary to track the
      object; if ``.exists()`` is False, raise a KeyError.
    :type strict: bool
    
    :keyword vars: Provide a custom Configuration class for command line options.
    :type vars: instance of any 
      :class:`anadama2.cli.Configuration` class or None.
      
    :keyword version: The version of the workflow. This version will be used for
      the command line option ``--version``.
    :type version: str
    
    :keyword description: A description of the workflow. This description 
      will be used in the command line ``--help`` message.
    :type description: str
    
    :keyword remove_options: A list of options to remove
    :type remove_options: list
    
    :keyword document: Provide a custom Document class.
    :type: document: instance of any :class: `anadama2.Document' class or None.
    """


    def __init__(self, storage_backend=None, grid=None, strict=False,
                 vars=None, version=None, description=None, remove_options=None,
                 document=None, cli=True):
        self.task_counter = itertools.count()
        self.dag = nx.DiGraph()
        #: tasks is a :class:`anadama2.taskcontainer.TaskContainer`
        #: filled with objects of type :class:`anadama2.Task`. This
        #: list is populated as new tasks are added via
        #: :meth:`anadama2.workflow.Workflow.add_task` and
        #: :meth:`anadama2.workflow.Workflow.do`
        self.tasks = TaskContainer()
        #: task_results is a list of objects of type 
        #: :class:`anadama2.runners.TaskResult`. This list is populated
        #: only after tasks have been run with
        #: :meth:`anadama2.workflow.Workflow.go`.
        self.task_results = list()
        self._depidx = tracked.DependencyIndex()
        if grid:
            self.grid = grid
            self.grid_set = True
        else:
            self.grid = _grid.Dummy()
            self.grid_set = False
        self.strict = strict
        self.vars = vars or Configuration(description=description,
            version=version, defaults=True, remove_options=remove_options, prompt_user=cli)
        
        if document:
            self.document=document
        else:
            self.document=PweaveDocument()

        self._backend=None
        if storage_backend:
            self._backend = storage_backend

        # get the temp directory location
        self.tmpdir=self.vars.get("output")
        if self.tmpdir is None or tracked.s3_folder(self.tmpdir):
            # if no output folder is provided, then write to the current working directory
            self.tmpdir = os.getcwd()

        logger.debug("Instantiated run context")
        
    def _get_grid(self):
        """Return the grid instance. First check if a grid instance has already 
        been set or was provided initially. If not, set the grid based on the 
        command line options. A dummy grid is set if the option to run is not
        provided on the command line. In this case, all tasks will run locally."""
        
        # if the grid has already been set by the user, then return
        if self.grid_set:
            return self.grid
        
        # get the grid selection set on the command line
        grid_selection = self.vars.get("grid")
        grid_partition = self.vars.get("grid_partition")
        grid_jobs = self.vars.get("grid_jobs")
        grid_options = self.vars.get("grid_options")
        grid_environment = self.vars.get("grid_environment")
        grid_benchmark_setting = True if self.vars.get("grid_benchmark") == "on" else False
        
        # set the grid instance based on the user option
        # if no grid jobs are selected, then no grid will be used
        if grid_jobs == 0:
            grid = _grid.Dummy()
        elif grid_selection == "slurm":
            # get the temp output folder for the slurm scripts and stdout/stderr files
            tmpdir = os.path.join(self.tmpdir, "slurm_files")
            grid = Slurm(partition=grid_partition, tmpdir=tmpdir, benchmark_on = grid_benchmark_setting,
                options=grid_options, environment=grid_environment)
        elif grid_selection == "sge":
            # get the temp output folder for the sge scripts and stdout/stderr files
            tmpdir = os.path.join(self.tmpdir, "sge_files")
            grid = SGE(partition=grid_partition, tmpdir=tmpdir, benchmark_on = grid_benchmark_setting,
                options=grid_options, environment=grid_environment)
        elif grid_selection == "aws":
            # get the temp output folder for the aws scripts and stdout/stderr files
            tmpdir = os.path.join(self.tmpdir, "aws_files")
            grid = AWS(partition=grid_partition, tmpdir=tmpdir)
        else:
            print("Grid selected ( "+grid_selection+" ) can not be found. Tasks will run locally.")
            grid = _grid.Dummy()
            
        self.grid=grid
        self.grid_set=True
        
        return self.grid
        

    def add_argument(self, name, **kwargs):
        """This function adds an argument to the configuration object provided
        to the workflow. Arguments can alternatively be added to the configuration
        object before it is provided to the workflow. See the ``add`` function
        documentation for your Configuration class, for more information. The 
        default configuration class is :class:`anadama2.cli.Configuration`."""
        
        self.vars.add(name, **kwargs)

    def parse_args(self):
        """Return the arguments parsed from the command line. Arguments are returned
        in the same format as calling argparse.parse_args(). An object with
        an attribute for each argument is returned. This custom object includes error
        reporting to aid the user in debugging when trying to get an argument
        that is not included in the list of command line arguments."""

        # return the arguments
        return self.vars.get_option_values()
    

    def get_input_files(self, extension=None, name=None):
        """Return the files in the input folder filtered with the extension
        or name if provided. The input folder default can be set in the workflow
        or it can be provided on the command line by the user.
        
        :keyword extension: Return input files with this extension
        :type extension: str
        :keyword name: Return input files with this name
        :type name: str

        :returns: A list of files
        """
        
        # get the contents of the input folder (with the full paths)
        vars_input = self.vars.get("input")
        if tracked.s3_folder(vars_input):
            import boto3
            client = boto3.client("s3")
            bucket = tracked.s3_bucket(vars_input)
            contents = [i['Key'] for i in client.list_objects(Bucket=bucket)['Contents']]
            input_files = [tracked.s3_build_path(bucket,file) for file in filter(lambda file: not file.endswith("/"), contents)]
        else:
            input = os.path.abspath(vars_input)
        
            input_folder_contents = map(lambda file: os.path.join(input, file), os.listdir(input))
            # filter out contents to only include files
            input_files = [item for item in input_folder_contents if os.path.isfile(item)]

        # if extension is set, then filter files
        if extension:
            input_files = list(filter(lambda file: file.endswith(extension), input_files))

        # if name is set, then filter files to only those with the exact name
        if name:
            input_files = list(filter(lambda file: os.path.basename(file) == name, input_files))
            
        return input_files
    
    def name_output_files(self, name, tag=None, extension=None, subfolder=None):
        """Return names of files in the output folder 
        use the name(s), tag, extension, and subfolder provided. 
        The output folder default can be set in the workflow or 
        it can be provided on the command line by the user.
        
        :parameter name: The name(s) of the output file(s)
        :type name: str or list
        :keyword tag: Add tag to the basename
        :type tag: str
        :keyword extension: Add extension to file name
        :type extension: str
        :keyword subfolder: Add the subfolder to the path to the file
        :type subfolder: str

        :returns: A list of file names
        """
        # if the name is a string, convert to list
        convert_to_string=False
        if not isinstance(name, list):
            name=[name]
            convert_to_string=True
       
        output_folder = self.vars.get("output") 
        if not tracked.s3_folder(output_folder):
            # get the output folder name (full path)
            output_folder = os.path.abspath(output_folder)

        # add the subfolder if provided
        if subfolder:
            output_folder = os.path.join(output_folder, subfolder)
            
        # Add tag and extension to file names
        output_files = [ fname.mangle(file, tag=tag, dir=output_folder, ext=extension) for file in name]
            
        # If the input was a single string, then convert output to single string
        if convert_to_string:
            output_files=output_files[0]
        
        return output_files

    def do(self, cmd, track_cmd=True, track_binaries=True):
        """Create and add a :class:`anadama2.Task` to the workflow using a
        convenient, shell-like syntax. 

        To explicitly mark task targets, wrap filenames within ``cmd``
        with ``[t:]``. Similarly, wrap dependencies with ``[d:]``. The
        literal ``[t:]`` and ``[d:]`` characters will be stripped out
        prior to execution by the shell.

        Below are some examples of using ``do``:

        .. code:: python

            from anadama2 import Workflow

            ctx = Workflow()
            ctx.do("wget -qO- checkip.dyndns.com > [t:my_ip.txt]")
            ctx.do(r"sed 's|.*Address: \(.*[0-9]\)<.*|\1|' [d:my_ip.txt] > [t:ip.txt]")
            ctx.do("whois $(cat [d:ip.txt]) > [t:whois.txt]")
            ctx.go()


        Variables from the workflow configuration can also be used
        inside ``cmd``. These are wrapped with ``[v:]``:

        .. code:: python

            from anadama2 import Workflow

            ctx = Workflow()
            ctx.do("wget -qO- checkip.dyndns.com > [v:output]/[t:my_ip.txt]")
            ctx.go()


        Modifiers inside the square brackets can be mixed and matched:

        .. code:: python

            from anadama2 import Workflow
            from anadama2.cli import Configuration

            ctx = Workflow(vars=Configuration().add("input", type="dir"))
            ctx.do("tar c [vd:input] | gzip -c > [t:output.tgz]")
            ctx.go()



        By default, changes made to ``cmd`` are tracked; any changes
        to ``cmd`` will cause this task to be rerun. Set ``track_cmd``
        to False to disable this behavior.

        Also by default, AnADAMA tries to discover pre-existing, small
        files in ``cmd`` and treat them as dependencies. This feature
        is intended to automatically track the scripts and binaries
        used in ``cmd``. Thus, this task will be re-run if any of the
        binaries or scripts change. Set ``track_binaries`` to False to
        disable this behavior.

        :param cmd: The shell command to add to the workflow. Wrap a
          target filename in ``[t:]`` and wrap a dependency filename
          in ``[d:]``. Variables from workflow configuration can be
          substituted into the command by wrapping the variable name
          in ``[v:]``. 
        :type cmd: str

        :keyword track_cmd: Set to False to not track changes to ``cmd``.
        :type track_cmd: bool
        
        :keyword track_binaries: Set to False to not discover files
          within ``cmd`` and treat them as dependencies.
        :type track_binaries: bool

        :returns: The :class:`anadama2.Task` just created

        """
        targs, ds = [], []

        def _repl(match):
            modifiers, name = match.groups()
            if 'v' in modifiers:
                name = self.vars.get(name) or _miss_exc(name)
            if 'd' in modifiers:
                ds.append(name)
            elif 't' in modifiers:
                targs.append(name)
            return str(name)

        sh_cmd = re.sub(r'\[([vdt]+):([^][]+)\]', _repl, cmd)
        if track_cmd:
            ns = os.path.abspath(tracked.Container.key(None))
            varname = "task_{}_command".format(len(self.tasks)+1)
            d = tracked.TrackedVariable(ns, varname, sh_cmd)
            ds.append(d)
        if track_binaries:
            to_preexist = []
            for binary in discover_binaries(cmd):
                to_preexist.append(binary)
                ds.append(binary)
            if to_preexist:
                self.already_exists(*to_preexist)

        return self.add_task(sh_cmd, depends=ds, targets=targs, name=sh_cmd,
                             interpret_deps_and_targs=False)


    def do_gridable(self, cmd, track_cmd=True, track_binaries=True, **gridopts):
        """Add a task to be launched on a grid computing system as specified
        in the ``grid`` option of
        :class:`anadama2.workflow.Workflow`. By default, this
        method is a synonym for
        :meth:`anadama2.workflow.Workflow.do`. Please see the
        ``add_task`` documentation for your powerup of choice
        e.g. :meth:`anadama2.slurm.Slurm.do` for information on
        options to provide to this method.
        """

        t = self.do(cmd, track_cmd, track_binaries)
        self._get_grid().do(t, **gridopts)
        return t
    
    def add_document(self, templates, depends=None, targets=None, vars=None, table_of_contents=None):
        """ Create and add a group of :class:`anadama2.Task` to the workflow. This
        task will create a document which will be the target(s) provided. The
        variables will be passed on to the template and be available when the
        document is generated from the template. The document class provided
        to the workflow will be used to create the document."""
        
        doc = copy.deepcopy(self.document)
        doc.__init__(templates, depends, targets, vars, table_of_contents)
        
        return self.add_task(actions=doc.create, depends=depends, targets=targets,
                      interpret_deps_and_targs=False, name="document")  
        
    def add_archive(self, depends, targets, archive_software=None, remove_log=None):
        """ Create an archive including the dependencies. Name it the target. This
        adds a :class:`anadama2.Task` to the workflow to create the archive. """

        # convert the depends and targets to lists if strings
        # update to absolute paths
        depends=sugar_list(depends)
        targets=[os.path.abspath(target) for target in sugar_list(targets)]
        
        # get the absolute path to the output folder
        if not self.vars.get("output") is None and not tracked.s3_folder(self.vars.get("output")):
            output_folder = os.path.abspath(self.vars.get("output"))
        else:
            output_folder = None
        
        # determine the archive software based on the target extension
        if archive_software is None:
            if targets[0].endswith(".zip"):
                archive_software = "zip"
            else:
                archive_software = "tar"
                
        # get the archive extension
        if archive_software == "zip":
            extension=".zip"
        else:
            extension=".tar"+targets[0].split("tar")[-1]
                
        # normalize a path to remove any extra separators included by the user
        # allows joining a folder ending with a separator with an extension as a target
        targets[0]=targets[0].replace(os.pathsep+extension,extension)
            
        # remove any tasks from the depends list of archive inputs
        archive_inputs=list(filter(lambda x: not isinstance(x,Task), depends))
        
        # if there is an output folder, change the depends to relative paths
        # so the full path is not included in the archive
        if output_folder is not None:
            archive_inputs=[os.path.relpath(file, start=output_folder) for file in archive_inputs]

        # check that the software is installed to create the archive
        try:
            output=subprocess.check_output([archive_software,"--help"])
        except EnvironmentError:
            raise EnvironmentError("Unable to archive document with software requested. Please install " +archive_software)
        
        # get the possible location of the db folder and the log
        document_db_folder="*"+backends.LOCAL_DB_FOLDER+"*"
        document_log=reporters.LOG_FILE_NAME
        
        # create the command always removing the database, if found
        # if the database is not found this will not issue an error it just will not be excluded
        if archive_software == "zip":
            command=["zip","-r",targets[0]]
            command+=archive_inputs
            command+=["-x",document_db_folder]
            if remove_log:
                command+=["-x",document_log]
        elif archive_software == "tar": 
            if targets[0].endswith(".tar.gz"):
                settings="czvf"
            elif targets[0].endswith(".tar"):
                settings="cvf"
            elif targets[0].endswith(".tar.bz2"):
                settings="cjvf"
            else:
                raise TypeError("Unknown archive software extension. Please use .tar.gz, .tar, or .tar.bz2 with tar software.")
                
            command=["tar",settings,targets[0]]
            command+=archive_inputs
            command+=["--exclude",document_db_folder]
            if remove_log:
                command+=["--exclude",document_log]
        else:
            raise TypeError("Unknown archive software provided. Please select zip or tar with .tar.gz or .tar or .tar.bz2")
        
        # move to the directory of the output folder to not include full paths in archives
        # if there is an output folder being used
        if output_folder is not None:
            cwd=os.getcwd()
            command=["cd",output_folder,";"]+command+[";","cd",cwd]
            
        command=" ".join(command)
        
        return self.add_task(actions=command, depends=depends, targets=targets,
                      interpret_deps_and_targs=False, name="archive")

    def add_task_group(self, actions=None, depends=None, targets=None,
                       name=[None], interpret_deps_and_targs=True, **kwargs):
        """Create and add a group of :class:`anadama2.Task` to the workflow. 
        This function will create a task for each set of depends and targets
        provided. The number of targets and dependencies should be the same.
        
        This function will call ``add_task`` for each task in the group. Please
        see the ``add_task`` documentation for more information."""
        
        task_group=[]
        for deps, targs, tname in zip(depends, targets, itertools.cycle(sugar_list(name))):
            task_group.append(self.add_task(actions, deps, targs, tname, interpret_deps_and_targs, **kwargs))
            
        return task_group
            
    def add_task_group_gridable(self, actions=None, depends=None, targets=None,
                       name=[None], interpret_deps_and_targs=True, **kwargs):
        """Create gridable tasks as a group."""
        
        task_group=[]
        for deps, targs, tname in zip(depends, targets, itertools.cycle(sugar_list(name))):
            task = self.add_task(actions, deps, targs, tname, interpret_deps_and_targs, **kwargs)
            task_group.append(task)
            self._get_grid().add_task(task, **kwargs)
        
        return task_group        

    def add_task(self, actions=None, depends=None, targets=None,
                 name=None, visible=True,
                 interpret_deps_and_targs=True, **kwargs):
        """Create and add a :class:`anadama2.Task` to the workflow.  This
        function can be used as a decorator to set a function as the
        sole action. 

        Extra keyword arguments can be used as formatting values
        similar to ``[depends[0]]``. See :func:`anadama2.helpers.parse_sh`
        
        :param actions: The actions to be performed to complete the
          task. Strings or lists of strings are interpreted as shell
          commands according to :func:`anadama2.helpers.parse_sh`. If given
          just a string or just a callable, this method treats it as a
          one-item list of the string or callable.
        :type actions: str or callable or list of str or
          list of callable

        :param depends: The dependencies of the task. The task must
          have these dependencies before executing the
          actions. Strings or lists of strings are interpreted as
          filenames and turned into objects of type
          :class:`anadama2.tracked.HugeTrackedFile`. If given just a string or just
          a :class:`anadama2.tracked.Base`, this method treats it as a
          one-item list of the argument provided.
        :type depends: str or :class:`anadama2.tracked.Base` or list of
          str or list of :class:`anadama2.tracked.Base`

        :param targets: The targets of the task. The task must produce
          these targets after executing the actions to be considered
          as "success". Strings or lists of strings are interpreted as
          filenames and turned into objects of type
          :class:`anadama2.tracked.HugeTrackedFile`. If given just a string or just
          a :class:`anadama2.tracked.Base`, this method treats it as a
          one-item list of the argument provided.
        :type targets: str or :class:`anadama2.tracked.Base` or list of
          str or list of :class:`anadama2.tracked.Base`

        :param name: A name for the task. Task names must be unique
          within a run context.
        :type name: str

        :keyword visible: Whether to show this task on the console. Set
          to ``False`` if it should only be in the debug log.
        :type visible: bool

        :keyword interpret_deps_and_targs: Should I use
          :func:`anadama2.helpers.parse_sh` to change
          ``[depends[0]]`` and ``[targets[0]]`` into the first item in
          depends and the first item in targets? Default is True 
        :type interpret_deps_and_targs: bool

        :returns: The :class:`anadama2.Task` just created

        """

        deps = _build_depends(depends)
        targs = _build_targets(targets)
        task_no = next(self.task_counter)
        if not actions: # must be a decorator
            def finish_add_task(fn):
                the_task = Task(name, [fn], deps, targs, task_no, bool(visible), [fn], kwargs, interpret_deps_and_targs)
                self._add_task(the_task)
            return finish_add_task
        else:
            # if any targets or depends generate temp files, create temp folder for task
            tracked_with_temp = list(filter(lambda x: x.temp_files(), deps+targs))
            tmpdir = os.path.join(self.tmpdir,"anadama2_temp_tracked")
            if visible and tracked_with_temp:
                if not os.path.isdir(tmpdir):
                    os.makedirs(tmpdir)
                tmpdir = tempfile.mkdtemp(dir=tmpdir)

                # set local temp locations for tracked items
                tmp = [tracked.try_set_local_path(x, tmpdir) for x in tracked_with_temp]

            acts = build_actions(actions, deps, targs, visible, kwargs, 
                                  use_parse_sh=interpret_deps_and_targs)

            the_task = Task(name, acts, deps, targs, task_no, bool(visible), actions, kwargs, interpret_deps_and_targs)
            self._add_task(the_task)
            return the_task


    def add_task_gridable(self, actions=None, depends=None, targets=None,
                      name=None, interpret_deps_and_targs=True, **gridopts):
        """Add a task to be launched on a grid computing system as specified
        in the ``grid`` option of
        :class:`anadama2.workflow.Workflow`. By default, this
        method is a synonym for
        :meth:`anadama2.workflow.Workflow.add_task`. Please see the
        ``add_task`` documentation for your powerup of choice
        e.g. :meth:`anadama2.grid.slurm.Slurm.add_task` for information on
        options to provide to this method.

        """
        if not actions: # must be a decorator
            def finish_grid_add_task(fn):
                t = self.add_task([fn], depends, targets, name)
                self._add_task(t)
                self.grid.add_task(t, **gridopts)
            return finish_grid_add_task
        else:
            t = self.add_task(actions, depends, targets, name,
                              interpret_deps_and_targs, **gridopts)
            self._get_grid().add_task(t, **gridopts)
            return t


    def already_exists(self, *depends):
        """Declare a dependency as pre-existing. That means that no task
        creates these dependencies; they're already there before any
        tasks run.

        .. note::

            If you have a list or other iterable containing the
            dependencies that already exist, you can declare them all
            like so ``ctx.already_exists(*my_bunch_of_deps)``.

        :param \*depends: One or many dependencies to mark as pre-existing.
        :type \*depends: any argument recognized by :func:`anadama2.tracked.auto`

        """

        self.add_task(noop, targets=list(map(tracked.auto, depends)),
                      name="Track pre-existing dependencies", visible=False)


    def go(self, skip_nothing=False, quit_early=False, runner=None,
           reporter=None, jobs=None, grid_jobs=None,
           until_task=None, exclude_task=None, target=None,
           exclude_target=None, dry_run=False):
        """Kick off execution of all previously configured tasks. 

        :keyword skip_nothing: Skip no tasks, even if you could.
        :type skip_nothing: bool
        
        :keyword quit_early: If any tasks fail, stop all execution
          immediately. If set to ``False`` (the default), children of
          failed tasks are *not* executed but children of successful
          or skipped tasks *are* executed: basically, keep going until
          you run out of tasks to execute.

        :keyword runner: The tasks to execute are passed to this
          object for execution.  For a list of runners that come
          bundled with anadama, see :mod:`anadama2.runners`. Passing
          ``None`` (the default) uses the default runner from
          :func:`anadama2.runners.default`.
        :type runner: instance of any
          :class:`anadama2.runners.BaseRunner` subclass or None.

        :keyword reporter: As task execution proceeds, events are
          dispatched to this object for reporting purposes. For more
          information of the reporters bundled with anadama, see
          :mod:`anadama2.reporters`. Passing ``None`` (the default)
          uses the default reporter from
          :func:`anadama2.reporters.default`.
        :type reporter: instance of any
          :class:`anadama2.reporters.BaseReporter` subclass or None.

        :keyword jobs: The number of tasks to execute in
          parallel. This option is ignored when a custom runner is
          used with the ``runner`` keyword.
        :type jobs: int

        :keyword grid_jobs: The number of tasks to submit to the
          grid in parallel. This option is ignored when a custom
          runner is used with the ``runner`` keyword. This option is
          also a synonym for ``jobs`` if the context has no grid
          powerup.
        :type grid_jobs: int

        :keyword until_task: Stop after running the named task. Can
          refer to the end task by task number or task name.
        :type until_task: int or str

        :keyword exclude_task: Don't execute this task or any of its
          children. Can refer to the task by task number or task name.
        :type exclude_task: int or str

        :keyword target: Execute the necessary tasks to produce this
          target. If ``target`` contains ``[``, ``*``, or ``?``, it is
          treated as a pattern and used to match multiple targets.
        :type target: str

        :keyword exclude_target: Don't execute any tasks that will
          produce this target. If ``target`` contains ``[``, ``*``, or
          ``?``, it is treated as a pattern and used to match multiple
          targets.
        :type exclude_target: str

        :keyword dry_run: Don't execute any actions, just say that you
          did.
        :type dry_run: bool

        """
        skip_nothing   = skip_nothing   or self.vars.get("skip_nothing")
        quit_early     = quit_early     or self.vars.get("quit_early")
        jobs           = jobs           or self.vars.get("jobs")
        grid_jobs      = grid_jobs      or self.vars.get("grid_jobs")
        until_task     = until_task     or self.vars.get("until_task")
        exclude_task   = exclude_task   or self.vars.get("exclude_task")
        target         = target         or self.vars.get("target")
        exclude_target = exclude_target or self.vars.get("exclude_target")
        dry_run        = dry_run        or self.vars.get("dry_run")

        self.completed_tasks = set()
        self.failed_tasks = set()
        self.task_results = [None for _ in range(len(self.tasks))]
        self._reporter = reporter or reporters.default(self.vars.get("output"),self.vars.get("log_level"))
        self._reporter.started(self)
        
        # if the backend is not set, then set to default
        if not self._backend:
            self._backend = backends.default(self.vars.get("output"))

        _runner = runner or self.grid.runner(self, jobs, grid_jobs)
        if dry_run:
            _runner = runners.DryRunner(self)
        _runner.quit_early = quit_early
        logger.debug("Sorting task_nos by network topology")
        task_idxs = nx.algorithms.dag.topological_sort(self.dag, reverse=True)
        logger.debug("Sorting complete")
        keep, drop = set(), set()
        if until_task:
            for task_name_or_no in sugar_list(until_task):
                for t in self._taskmatch(task_name_or_no):
                    keep = keep.union(allparents(self.dag, t.task_no))
        if exclude_task:
            for task_name_or_no in sugar_list(exclude_task):
                for t in self._taskmatch(task_name_or_no):
                    drop = drop.union(allchildren(self.dag, t.task_no))
        if target:
            for name_or_pattern in sugar_list(target):
                keep = self._targetmatch(keep, name_or_pattern, allparents)
        if exclude_target:
            for name_or_pattern in sugar_list(exclude_target):
                drop = self._targetmatch(drop, name_or_pattern, allchildren)
        if not keep:
            keep = set(task_idxs)
        task_idxs = list(filter((keep-drop).__contains__, task_idxs))
        if not skip_nothing:
            task_idxs = self._filter_skipped_tasks(task_idxs)
        task_idxs = deque(task_idxs)

        _runner.run_tasks(task_idxs)
        self._handle_finished()


    def _import(self, task_dict):
        keys_to_keep = ("actions", "depends", "targets",
                        "name", "interpret_deps_and_targs")
        return self.add_task(**keepkeys(task_dict, keys_to_keep))

    _ = _import


    def _handle_task_result(self, result):
        if result.task_no is not None:
            self.task_results[result.task_no] = result
        if result.error:
            self.failed_tasks.add(result.task_no)
            self._reporter.task_failed(result)
        else:
            self._backend.save(result.dep_keys, result.dep_compares)
            self.completed_tasks.add(result.task_no)
            self._reporter.task_completed(result)
            pxdeps = [ d for d in self.tasks[result.task_no].depends
                       if not istask(d) and d not in self._depidx ]
            if pxdeps:
                self._backend.save([d.name for d in pxdeps], 
                                   [list(d.compare()) for d in pxdeps])
                    


    def _handle_finished(self):
        self._reporter.finished()
        if self.failed_tasks:
            raise RunFailed()


    def _filter_skipped_tasks(self, task_idxs):
        should_run, idxs = dichotomize(task_idxs, self._always_rerun)
        should_run = set(should_run)
        for dep, idxs_set in self._aggregate_deps(idxs):
            if tracked.any_different([dep], self._backend):
                for idx in idxs_set:
                    logger.debug("Can't skip task %i because of dep change",
                                 idx)
                    should_run.add(idx)
        while idxs:
            idx = idxs.pop()
            if idx in should_run:
                continue
            for parent_idx in self.dag.predecessors(idx):
                if parent_idx in should_run:
                    should_run.add(idx)
                    logger.debug("Can't skip %i because it depends "
                                 "on task %i, which will be rerun",
                                 idx, parent_idx)

        to_run, skipped = dichotomize(task_idxs, should_run.__contains__)
        for idx in skipped:
            self._handle_task_skipped(idx)
        return to_run


    def _always_rerun(self, task_no):
        task = self.tasks[task_no]
        if not task.targets and not task.depends:
            logger.debug("Can't skip task %i because it "
                          "has no targets or depends", task_no)
            return True
        return False


    def _aggregate_deps(self, idxs):
        grp = defaultdict(set)
        for idx in idxs:
            task = self.tasks[idx]
            for dep in itertools.chain(task.depends, task.targets):
                if istask(dep):
                    continue
                grp[dep].add(idx)
                    
        return six.iteritems(grp)
        

    def _handle_task_started(self, task_no):
        self._reporter.task_started(task_no)


    def _handle_task_skipped(self, task_no):
        self.completed_tasks.add(task_no)
        self._reporter.task_skipped(task_no)

    def _add_task(self, task):
        """Actually add a task to the internal dependency data structure"""
        
        self.tasks.append(task)
        self.dag.add_node(task.task_no)
        for dep in task.depends:
            if istask(dep):
                self.dag.add_edge(dep.task_no, task.task_no)
                continue
            if dep in self._depidx:
                pass
            elif dep.must_preexist == False: 
                continue
            elif not self.strict and dep.exists():
                self.already_exists(dep)
            try:
                parent_task = self._depidx[dep]
            except KeyError:
                self._handle_nosuchdep(dep, task)
            else:
                # check to see if the dependency exists but doesn't
                # link to a task. This would happen if someone defined
                # a preexisting dependency
                if parent_task is not None:
                    self.dag.add_edge(parent_task.task_no, task.task_no)
        for targ in task.targets: 
            # add targets to the DependencyIndex after looking up
            # dependencies for the current task. Hopefully this avoids
            # circular references
            self._depidx.link(targ, task)
                

    def _handle_nosuchdep(self, dep, task):
        self.tasks.pop()
        self.dag.remove_node(task.task_no)
        msg = "Unable to find dependency `{}' of type `{}'. "
        msg = msg.format(str(dep), type(dep))
        alldeps = itertools.chain.from_iterable(
            [list(t.depends) + list(t.targets) for t in self.tasks]
        )
        try:
            closest = matcher.find_match(dep, alldeps, key=attrgetter("name"))
        except:
            raise KeyError(msg)
        msg += "Perhaps you meant `{}' of type `{}'?"
        raise KeyError(msg.format(str(closest), type(closest)))


    @property
    def _alltargets(self):
        return iter( (targ.name, task.task_no) for task in self.tasks
                     for targ in task.targets )

    def _targetmatch(self, s, name_or_pattern, hier, try_cwd=False):
        # try the target with the full path to the current working directory
        if try_cwd:
            name_or_pattern=os.path.join(os.getcwd(), name_or_pattern)

        if re.search(r'[?*\[]', name_or_pattern):
            regex = re.compile(fnmatch.translate(name_or_pattern))
            matches = [ no for name, no in self._alltargets
                        if regex.match(name) ]
            ret = set( sibling for match in matches
                       for sibling in hier(self.dag, match) )
            if not ret:
                msg = "Pattern {} matched no targets."
                if try_cwd:
                    msg+=" Tried without path and then with expected path. Please provide the full path."
                elif not name_or_pattern.startswith(os.sep):
                    return self._targetmatch(s, name_or_pattern, hier, try_cwd=True)
                raise ValueError(msg.format(name_or_pattern))
        else:
            try:
                match = next( iter(no for name, no in self._alltargets
                                   if name_or_pattern == name) )
            except StopIteration:
                msg = "Unable to find target {}.".format(name_or_pattern)
                if try_cwd:
                    msg+=" Tried without path and then with expected path. Please provide the full path."
                elif not name_or_pattern.startswith(os.sep):
                    return self._targetmatch(s, name_or_pattern, hier, try_cwd=True)
                raise ValueError(msg)
            ret = set(hier(self.dag, match))
        return s.union(ret)

    def _taskmatch(self, task_name_or_number):
        # Find the tasks that match the name or number provided
        # Multiple tasks can share the same name but all will have unique numbers
        
        taskset=[]
        for task in self.tasks:
            if task.name == task_name_or_number:
                taskset.append(task)
            elif task.task_no == task_name_or_number:
                taskset.append(task)
                
        return taskset

def _build_depends(depends):
    depends = filter(None, sugar_list(depends))
    return list(map(tracked.auto, depends))


def _build_targets(targets):
    targets = filter(None, sugar_list(targets))
    ret = list()
    for targ in targets:
        if istask(targ):
            raise ValueError("Can't make a task a target")
        ret.append(tracked.auto(targ))
    return ret


def _miss_exc(name):
    msg = "Unable to find configuration variable `{}'".format(name)
    raise Exception(msg)


def discover_binaries(s):
    """Search through string ``s`` and find all existing files smaller
    than 10MB. Return those files as a list of objects of type
    :class:`anadama2.tracked.TrackedExecutable`.
    """

    ds = list()
    for term in shlex.split(s):
        if not os.path.exists(term):
            term = find_on_path(term)
        if not term:
            continue
        if os.path.isdir(term):
            # don't want directories
            continue
        if not os.access(term, os.F_OK | os.X_OK):
            # doesn't exist or can't execute
            continue
        try:
            dep = tracked.TrackedExecutable(term)
        except ValueError:
            continue
        if os.stat(dep.name).st_size < 1<<20:
            ds.append(dep)

    return ds


def allchildren(dag, task_no):
    kids = map(second, dfs_edges(dag, task_no))
    return itertools.chain([task_no], kids)


def allparents(dag, task_no):
    seen = set()
    to_check = deque([task_no])
    while to_check:
        idx = to_check.popleft()
        if idx in seen:
            continue
        seen.add(idx)
        to_check.extend(dag.predecessors(idx))
    return seen



# -*- coding: utf-8 -*-
from __future__ import print_function
import time
import logging
import itertools
import traceback
import multiprocessing
import pickle as pickle
from collections import namedtuple

import six
from six.moves import queue
import cloudpickle

from .util import istask
from . import tracked
from .helpers import apply_sh

logger = logging.getLogger(__name__)


class TaskResult(namedtuple(
        "TaskResult", ["task_no", "error", "dep_keys", "dep_compares"])):
    """The result of a task's execution.

    :param task_no: The number of the task that produced this result.
    :type: int

    :param error: The error generated by executing the task. None if
      the task was successful.
    :type: str or None

    :param dep_keys: The list of ``.name`` attributes from the objects
      of type :class:`anadama2.tracked.Base` associated with
      this task. These keys are used with the storage backend to save
      successful task results.

    :type dep_keys: list of str

    :param dep_compares: The list of the results of the ``compare()``
      method from the objects of type
      :class:`anadama2.tracked.Base` associated with this
      task. These values are used with the storage backend to save
      successful task results.

    :type dep_keys: list of list of str

    """
    pass



class TaskFailed(Exception):
    def __init__(self, msg, task_no):
        self.task_no = task_no
        super(TaskFailed, self).__init__(msg)


class BaseRunner(object):
    def __init__(self, run_context):
        self.ctx = run_context
        self.quit_early = False

    def run_tasks(self, task_idx_deque):
        raise NotImplementedError()


class DryRunner(BaseRunner):

    _typemap = {
        tracked.TrackedString: "String",
        tracked.TrackedVariable: "Variable",
        tracked.TrackedFile: "File",
        tracked.HugeTrackedFile: "Big File",
        tracked.TrackedDirectory: "Directory",
        tracked.TrackedFilePattern: "File Pattern",
        tracked.TrackedExecutable: "Executable",
        tracked.TrackedFunction: "Python Function",
    }

    sublist_template="  - {} ({})"

    def run_tasks(self, task_idx_deque):
        # order the tasks by number
        tasks_by_number={}
        for index in task_idx_deque:
            number=self.ctx.tasks[index].task_no
            try:
                number=int(number)
            except (ValueError, TypeError):
                pass
                
            tasks_by_number[number]=self.ctx.tasks[index]
        # print the tasks from first to last
        for number, task in sorted(tasks_by_number.items()):
            six.print_("{} - {}".format(number, task.name))
            six.print_("  Dependencies ({})".format(len(task.depends)))
            for dep in task.depends:
                six.print_(self._depformat(dep))
            six.print_("  Targets ({})".format(len(task.targets)))
            for dep in task.targets:
                six.print_(self._depformat(dep))
            six.print_("  Actions ({})".format(len(task.actions)))
            for action in task.actions:
                six.print_(self._actionformat(action))
                    
            six.print_("------------------")


    def _depformat(self, d):
        if istask(d):
            return "  - Task {} - {}".format(d.task_no, d.name)
        else:
            t = type(d)
            desc = self._typemap.get(t, str(t))
            return self.sublist_template.format(d.name, desc)
    
    def _actionformat(self, action):
        if six.callable(action):
            return self.sublist_template.format(action.__name__,"function")
        else:
            return self.sublist_template.format(action,"command")



class SerialLocalRunner(BaseRunner):

    def run_tasks(self, task_idx_deque):
        total = len(task_idx_deque)
        logger.debug("Running %i tasks locally and serially", total)
        while task_idx_deque:
            idx = task_idx_deque.pop()

            parents = set(self.ctx.dag.predecessors(idx))
            failed_parents = parents.intersection(self.ctx.failed_tasks)
            if failed_parents:
                self.ctx._handle_task_result(
                    parent_failed_result(idx, next(iter(failed_parents))))
                continue

            self.ctx._handle_task_started(idx)
            self.ctx._reporter.task_running(idx)
            self.ctx._reporter.task_command(idx)
            result = _run_task_locally(self.ctx.tasks[idx])
            self.ctx._handle_task_result(result)

            if self.quit_early and bool(result.error):
                logger.debug("Quitting early")
                break


def worker_run_loop(work_q, result_q, run_task, reporter=None, lock=None):
    logger.debug("Starting worker")
    while True:
        try:
            logger.debug("Getting work")
            pkl, extra = work_q.get()
            logger.debug("Got work")
        except IOError as e:
            logger.debug("Received IOError (%s) errno %s from work_q",
                              e.message, e.errno)
            break
        except EOFError:
            logger.debug("Received EOFError from work_q")
            break
        if type(pkl) is dict and pkl.get("stop", False):
            logger.debug("Received sentinel, stopping")
            break
        try:
            logger.debug("Deserializing task")
            task = pickle.loads(pkl)
            logger.debug("Task deserialized")
        except Exception as e:
            result_q.put_nowait(exception_result(e))
            logger.debug("Failed to deserialize task")
            continue
        logger.debug("Running task %s with %s", task.task_no, run_task)
        if reporter:
            if lock is not None:
                lock.acquire()
            reporter.task_running(task.task_no)
            reporter.task_command(task.task_no)
            if lock is not None:
                lock.release()
        result = run_task(task, extra)
        logger.debug("Finished running task; "
                          "putting results on result_q")
        result_q.put_nowait(result)
        logger.debug("Result put on result_q. Back to get more work.")
    

def _run_task_locally(task, extra=None):
    # convert any command strings to sh actions before running
    for i, action_func in enumerate(apply_sh(task.actions)):
        # check if any files need to be copied to local paths
        if task.visible:
            tracked.download_files_if_needed(task.depends)
            tracked.create_temp_folders_if_needed(task.targets)
        logger.debug("Executing task %i action %i", task.task_no, i)
        try:
            action_func(task)
        except Exception:
            msg = ("Error executing action {}. "
                   "Original Exception: \n{}")
            return exception_result(
                TaskFailed(msg.format(i, traceback.format_exc()), task.task_no)
                )
        # upload files if need (not needed for invisible file tracking tasks)
        if task.visible:
            tracked.upload_files_if_needed(task.targets)
        logger.debug("Completed executing task %i action %i", task.task_no, i)

    return _get_task_result(task)

def _get_task_result(task):
    targ_keys, targ_compares = list(), list()
    for target in task.targets:
        targ_keys.append(target.name)
        try:
            targ_compares.append(list(target.compare()))
        except (Exception, EnvironmentError):
            msg = "Failed to produce target `{}'. Original exception: {}"
            return exception_result(
                TaskFailed(msg.format(target, traceback.format_exc()),
                           task.task_no)
                )
            
    return TaskResult(task.task_no, None, targ_keys, targ_compares)

    
class ParallelLocalWorker(multiprocessing.Process):
            
    def __init__(self, work_q, result_q, lock, reporter):
        super(ParallelLocalWorker, self).__init__()
        self.logger = logger
        self.work_q = work_q
        self.result_q = result_q
        self.lock = lock
        self.reporter = reporter


    @staticmethod
    def appropriate_q_class(*args, **kwargs):
        return multiprocessing.Queue(*args, **kwargs)
    
    @staticmethod
    def appropriate_lock():
        return multiprocessing.Lock()

    def run(self):
        return worker_run_loop(self.work_q, self.result_q, _run_task_locally, self.reporter, self.lock)


class ParallelLocalRunner(BaseRunner):

    def __init__(self, run_context, jobs):
        super(ParallelLocalRunner, self).__init__(run_context)
        self.work_q = multiprocessing.Queue()
        self.result_q = multiprocessing.Queue()
        self.lock = multiprocessing.Lock()
        self.reporter = run_context._reporter
        self.workers = [ ParallelLocalWorker(self.work_q, self.result_q, self.lock, self.reporter)
                         for _ in range(jobs) ]
        self.started = False


    def run_tasks(self, task_idx_deque):
        self.task_idx_deque = task_idx_deque
        logger.debug("Running %i tasks in parallel with %i workers locally",
                      len(task_idx_deque), len(self.workers))
        
        self.n_to_do = len(task_idx_deque)
        while True:
            self._fill_work_q()
            logger.debug("Tasks left to do: %s", self.n_to_do)
            if self.n_to_do <= 0:
                logger.debug("No new tasks added to work queues and"
                             " the number of completed tasks equals the"
                             " number of tasks to do. Job's done")
                break
            try:
                result = self.result_q.get()
            except (SystemExit, KeyboardInterrupt):
                logger.info("Terminating due to SystemExit or Ctrl-C")
                self.terminate()
                raise
            except Exception as e:
                logger.error("Terminating due to unhandled exception")
                logger.exception(e)
                self.terminate()
                raise
            else:
                self.n_to_do -= 1
                self.ctx._handle_task_result(result)
            if self.quit_early and result.error:
                logger.debug("Quitting early.")
                self.terminate()
                break

        self.cleanup()


    def _fill_work_q(self):
        logger.debug("Filling work_q")
        for _ in range(len(self.task_idx_deque)):
            idx = self.task_idx_deque.pop()
            parents = set(self.ctx.dag.predecessors(idx))
            failed_parents = parents.intersection(self.ctx.failed_tasks)
            if failed_parents:
                self.ctx._handle_task_result(
                    parent_failed_result(idx, failed_parents.pop()))
                self.n_to_do -= 1
                continue
            elif parents.difference(self.ctx.completed_tasks):
                # has undone parents, come back again later
                self.task_idx_deque.appendleft(idx)
                continue
            try:
                pkl = cloudpickle.dumps(self.ctx.tasks[idx])
            except Exception as e:
                msg = ("Unable to serialize task `{}'. "
                       "Original error was `{}'.")
                raise ValueError(msg.format(self.ctx.tasks[idx], e))
            logger.debug("Adding task %i to work_q", idx)
            self.ctx._handle_task_started(idx)
            self.work_q.put((pkl, None))
            logger.debug("Added task %i to work_q", idx)
        if not self.started:
            logger.debug("Starting up workers")
            for w in self.workers:
                w.start()
            self.started = True


    def terminate(self):
        logger.debug("Terminating all workers")
        self.work_q._rlock.acquire()
        logger.debug("got work_q readlock")
        while self.work_q._reader.poll():
            logger.debug("draining work_q")
            try:
                self.work_q._reader.recv()
            except EOFError:
                break
            time.sleep(0)
        for worker in self.workers:
            logger.debug("terminating worker %s", worker)
            worker.terminate()
        for worker in self.workers:
            logger.debug("joining worker %s", worker)
            worker.join()
        logger.debug("releasing readlock")
        self.work_q._rlock.release()
        logger.debug("termination complete")


    def cleanup(self):
        logger.debug("cleaning up parallellocalrunner")
        for w in self.workers:
            logger.debug("giving stop sentinel to worker %s", w)
            self.work_q.put(({"stop": True}, None))
        for w in self.workers:
            logger.debug("joining worker %s", w)
            w.join()
        logger.debug("successfully cleaned up parallellocalrunner")


class GridRunner(BaseRunner):

    def __init__(self, workflow):
        super(GridRunner, self).__init__(workflow)
        self._worker_config = dict()
        self._worker_qs = dict()
        self.routes = dict() # task_no -> (worker_type_name, extra_args)
        self.workers = list()
        self.started = False
        self.default_worker = None


    def add_worker(self, worker_class, name,
                   rate=1, default=False):
        self._worker_config[name] = (worker_class, rate)
        if default:
            self.default_worker = name


    def run_tasks(self, task_idx_deque):
        self.task_idx_deque = task_idx_deque
        if not self.workers:
            self._init_workers()

        logger.debug("Running %i tasks in parallel with %i workers"
                      " using the grid",
                     len(task_idx_deque), len(self.workers))
        self.n_to_do = len(task_idx_deque)
        while True:
            self._fill_work_qs()
            logger.debug("Tasks left to do: %s", self.n_to_do)
            if self.n_to_do <= 0:
                break
            try:
                result = self._get_result()
            except (SystemExit, KeyboardInterrupt):
                logger.info("Terminating due to SystemExit or Ctrl-C")
                self.terminate()
                raise
            except Exception as e:
                logger.error("Terminating due to unhandled exception")
                logger.exception(e)
                self.terminate()
                raise
            else:
                self.n_to_do -= 1
                self.ctx._handle_task_result(result)
            if self.quit_early and result.error:
                logger.debug("Quitting early.")
                self.terminate()
                break

        self.cleanup()


    def terminate(self):
        for name, (work_q, _) in self._worker_qs.items():
            if hasattr(work_q, "_rlock"):
                self._terminate_mpq(work_q, name)
            elif hasattr(work_q, "mutex"):
                self._terminate_qq(work_q, name)
            else:
                raise Exception
        for worker in self.workers:
            worker.join()


    def cleanup(self):
        for name, (_, n_procs) in self._worker_config.items():
            for _ in range(n_procs):
                self._worker_qs[name][0].put(({"stop": True}, None))
        for w in self.workers:
            w.join()


    def route(self, task_no):
        if task_no in self.routes:
            return self.routes[task_no]
        elif self.default_worker is not None:
            return self.default_worker, None
        else:
            msg = ("GridRunner tried to run task {} but has no "
                   "runner to run it and no default runner is defined")
            raise ValueError(msg.format(task_no))


    def _fill_work_qs(self):
        logger.debug("Filling work_qs")
        for _ in range(len(self.task_idx_deque)):
            idx = self._get_next_task()
            if idx is None:
                continue
            try:
                pkl = cloudpickle.dumps(self.ctx.tasks[idx])
            except Exception as e:
                msg = ("Unable to serialize task `{}'. "
                       "Original error was `{}'.")
                raise ValueError(msg.format(self.ctx.tasks[idx], e))
            name, extra = self.route(idx)
            logger.debug("Adding task %i to `%s' work_q", idx, name)
            self._worker_qs[name][0].put((pkl, extra))
            self.ctx._handle_task_started(idx)
            logger.debug("Added task %i to `%s' work_q", idx, name)
        if not self.started:
            logger.debug("Starting up workers")
            for w in self.workers:
                logger.debug("Starting worker %s", w)
                w.start()
            self.started = True


    def _init_workers(self):
        threads, procs = list(), list()
        for name, (worker_cls, n_procs) in self._worker_config.items():
            work_q = worker_cls.appropriate_q_class()
            result_q = worker_cls.appropriate_q_class()
            lock = worker_cls.appropriate_lock()
            self._worker_qs[name] = (work_q, result_q)
            isproc = issubclass(worker_cls, multiprocessing.Process) 
            l = procs if isproc else threads
            for _ in range(n_procs):
                l.append(worker_cls(work_q, result_q, lock, self.ctx._reporter))
        self.workers = procs+threads # http://stackoverflow.com/a/13115499
        self._qcycle = itertools.cycle(val[1]
                                       for val in self._worker_qs.values())


    def _get_next_task(self):
        idx = self.task_idx_deque.pop()
        parents = set(self.ctx.dag.predecessors(idx))
        failed_parents = parents.intersection(self.ctx.failed_tasks)
        if failed_parents:
            self.ctx._handle_task_result(
                parent_failed_result(idx, next(iter(failed_parents)))
                )
            self.n_to_do -= 1
            return None
        elif parents.difference(self.ctx.completed_tasks):
            # has undone parents, come back again later
            self.task_idx_deque.appendleft(idx)
            return None
        if idx is None:
            raise Exception
        return idx


    def _get_result(self):
        while True:
            for _ in range(len(self.workers)):
                try:
                    ret = next(self._qcycle).get(False)
                except queue.Empty:
                    continue
                return ret
            time.sleep(0.05)


    def _terminate_mpq(self, q, name):
        q._rlock.acquire()
        while q._reader.poll():
            try:
                q._reader.recv()
            except EOFError:
                break
            time.sleep(0)
        worker_type = self._worker_config[name][0]
        for worker in self.workers:
            if isinstance(worker, worker_type):
                worker.terminate()
        q._rlock.release()


    def _terminate_qq(self, q, name):
        q.mutex.acquire()
        while q.queue:
            q.queue.pop()
        worker_type = self._worker_config[name][0]
        for worker in self.workers:
            if isinstance(worker, worker_type):
                worker.join()
        q.mutex.release()
        


def default(run_context, jobs):
    if jobs < 2:
        return SerialLocalRunner(run_context)
    else:
        return ParallelLocalRunner(run_context, jobs)


_current_grid_runner = None
def current_grid_runner(context):
    global _current_grid_runner
    if _current_grid_runner is None:
        _current_grid_runner = GridRunner(context)
    return _current_grid_runner

    
def exception_result(exc):
    return TaskResult(getattr(exc, "task_no", None), str(exc), None, None)


def parent_failed_result(idx, parent_idx):
    return TaskResult(
        idx, "Task failed because parent task `{}' failed".format(parent_idx),
        None, None)





# -*- coding: utf-8 -*-
import re
import os
import glob
from itertools import chain
from operator import itemgetter

from six.moves import map, filter

DEFAULT_DATA_DIR = "./"

third = itemgetter(2)

def parse(pattern_str, data_dir=DEFAULT_DATA_DIR):
    if pattern_str.startswith("glob:"):
        pattern = os.path.join(data_dir,
                               pattern_str.split("glob:", 1)[1])
        files = list(map(os.path.abspath, glob.glob(pattern)))
    elif pattern_str.startswith("re:"):
        matcher = lambda s: re.search(pattern_str.split("re:", 1)[1], s)
        allfiles = chain.from_iterable( map(third, os.walk(data_dir)) )
        files = filter(matcher, allfiles)
    elif ',' in pattern_str:
        files = pattern_str.split(',')
        nonexistent = [ f for f in files if not os.path.exists(f) ]
        if nonexistent:
            raise OSError("No such file or directory: "+", ".join(nonexistent))
    else:
        if not os.path.exists(pattern_str):
            raise OSError("No such file or directory: "+pattern_str)
        files = [pattern_str]

    if not files:
        raise ValueError("The pattern matched no files")
    return files



# -*- coding: utf-8 -*-
from anadama2 import Workflow

ctx = Workflow(remove_options=["input","output"])
ctx.do("wget -qO- checkip.dyndns.com > [t:my_ip.txt]")
ctx.do(r"sed 's|.*Address: \(.*[0-9]\)<.*|\1|' [d:my_ip.txt] > [t:ip.txt]")
ctx.do("whois $(cat [d:ip.txt]) > [t:whois.txt]")
ctx.go()



# -*- coding: utf-8 -*-
import six

class Task(object):
    """A unit of work. 
    
    :param name: The task name; must be unique to all tasks within a workflow.
    :type name: str or unicode
    
    :param actions: The actions to execute; do these and the work is done.
    :type actions: list of callable
    
    :param depends: The list of dependencies. 
    :type depends: list of :class:`anadama2.tracked.Base`
    
    :param targets: The list of targets. The task must produce all of
      these to be a successfully complete task.
    :type targets: list of :class:`anadama2.tracked.Base`
    
    :param task_no: The unique task number. Ordered by declaration,
      not execution.
    :type task_no: int

    :param visible: Whether the task should appear in the console or
      not. These tasks do not generate targets (only track).
    :type visible: bool

    :param actions_raw: The raw actions from the user, unbuilt.
    :type actions: list of callable

    :param kwargs: Optional arguments used to build the actions.
    :type kwargs: dictionary

    :param use_parse_sh: Should parse actions resolving targets/dependencies.
    :type use_parse_sh: bool

    """
    
    def __init__(self, name, actions, depends, targets, task_no, visible, actions_raw, kwargs, use_parse_sh):
        # Set a default task number
        if task_no is None:
            self.task_no="NA"
        else:
            self.task_no=task_no
            
        # Set a default task name if not provided
        if name is None:
            self.name="Task"+str(task_no)
        else:
            self.name=name
            
        self.actions=actions
        self.depends=depends
        self.targets=targets
        self.visible=visible

        self.actions_raw=actions_raw
        self.kwargs=kwargs
        self.use_parse_sh=use_parse_sh
        
        # get a task description based on the actions or name
        if six.callable(actions[0]):
            # if the first action is a function, use the function name
            command=actions[0].__name__
        else:
            # if the first action is a command, use the executable name
            command=six.u(actions[0]).split(" ")[0]
            
        # if the task name is not set, then use the command name for the description
        if name is None:
            self.description=command
        else:
            self.description=six.u(name)
            
            
from .workflow import Workflow
from .document import PweaveDocument

Workflow # pyflakes




from anadama2 import Workflow

workflow = Workflow(remove_options=["input","output"])

# add a task to download the file
workflow.add_task(
    "wget ftp://public-ftp.hmpdacc.org/HMMCP/finalData/hmp1.v35.hq.otu.counts.bz2 -O [targets[0]]",
    targets="hmp1.v35.hq.otu.counts.bz2")

# add a task to decompress the file
workflow.add_task(
    "bzip2 -d < [depends[0]] > [targets[0]]",
    depends="hmp1.v35.hq.otu.counts.bz2",
    targets="hmp1.v35.hq.otu.counts")

def remove_end_tabs_function(task):
    with open(task.targets[0].name, 'w') as file_handle_out:
        for line in open(task.depends[0].name):
            file_handle_out.write(line.rstrip() + "\n")

# add a task with a function to remove the end tabs from the file
workflow.add_task(
    remove_end_tabs_function,
    depends="hmp1.v35.hq.otu.counts",
    targets="hmp1.v35.hq.otu.counts.notabs",
    name="remove_end_tabs")

workflow.go()





from anadama2 import Workflow

workflow = Workflow(remove_options=["input","output"])

downloads=["ftp://public-ftp.hmpdacc.org/HM16STR/by_sample/SRS011275.fsa.gz",
    "ftp://public-ftp.hmpdacc.org/HM16STR/by_sample/SRS011273.fsa.gz",
    "ftp://public-ftp.hmpdacc.org/HM16STR/by_sample/SRS011180.fsa.gz"]

for link in downloads:
    workflow.add_task(
        "wget -O [targets[0]] [args[0]]",
        targets=link.split("/")[-1],
        args=link)

workflow.go()




import anadama2.tracked
from anadama2 import Workflow

workflow = Workflow(remove_options=["input","output"])

# create a container class to track
container = anadama2.tracked.Container(a = 20)

# add a task that depends on the "a" variable in the container
task1=workflow.add_task(
    "echo [depends[0]] > [targets[0]]",
    depends=container.a, 
    targets="echo.txt",
    name="task1")

# add a task that depends on the targets of task1 
task2=workflow.add_task(
    "p=$(cat [depends[0]]); echo $p > [targets[0]]",
    depends=task1.targets[0],
    targets="echo2.txt",
    name="task2")

workflow.go()




# -*- coding: utf-8 -*-
import os
import sys

from setuptools import setup, find_packages
import distutils

# try to import urllib.request.urlretrieve for python3
try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

VERSION="0.6.0"
AUTHOR = "AnADAMA2 Development Team"
AUTHOR_EMAIL = "anadama-users@googlegroups.com"

# read requirements from file
requires_file=os.path.join(os.path.dirname(os.path.abspath(__file__)),"requirements.txt")
with open(requires_file) as file_handle:
    requires=[line.strip() for line in file_handle.readlines()]

if os.name != 'posix' or sys.version_info[0] > 2:
    try:
        requires.remove("subprocess32")
    except ValueError:
        pass

COUNTER_URL="http://bitbucket.org/biobakery/anadama2/downloads/counter.txt"

def download(url, download_file):
    """ Download a file from a url """

    try:
        print("Downloading "+url)
        file, headers = urlretrieve(url,download_file)
        # print final return to start new line of stdout
        print("\n")
    except EnvironmentError:
        print("WARNING: Unable to download "+url)

counter_file=os.path.basename(COUNTER_URL)
if not os.path.isfile(counter_file):
    print("Downloading counter file to track anadama2 downloads"+
    " since the global PyPI download stats are currently turned off.")
    download(COUNTER_URL,counter_file)

class SphinxBuild(distutils.cmd.Command):
    """ This is a custom command to build the API Docs"""
    description = "build the API docs"
    # add "=" to the option name to make it an argument instead of a flag
    user_options = [
        ('source=', None, 'the path to the sphinx source'),
        ('dest=', None, 'the path to the build destination')]

    def initialize_options(self):
        # set the source and dest to the defaults
        setup_location =  os.path.dirname(os.path.realpath(__file__))
        self.source = os.path.join(setup_location,"doc","source")
        self.dest = os.path.join(setup_location,"html")

    def finalize_options(self):
        # check that the directories provided exist
        if not os.path.isdir(self.source):
            sys.exit("The source directory does not exist,"+
                " please select another location: " + self.source)
        if not os.path.isdir(self.dest):
            sys.exit("The dest directory does not exist,"+
                " please select another location: " + self.dest)

    def run(self):
        # check for sphinx install
        import subprocess
        try:
            subprocess.check_output(["sphinx-build","--help"])
        except (EnvironmentError, subprocess.CalledProcessError):
            sys.exit("ERROR: Please install Sphinx ('pip install Sphinx')")

        # run the sphinx build command
        try:
            subprocess.check_call(["sphinx-build",self.source,self.dest])
        except (EnvironmentError, subprocess.CalledProcessError):
            sys.exit("ERROR: Unable to run sphinx build")

setup(
    name='anadama2',
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    version=VERSION,
    license="MIT",
    description="AnADAMA2: Another Automated Data Analysis Management Application 2",
    long_description="AnADAMA2 is the next generation of AnADAMA. AnADAMA is "+\
        "a tool to create reproducible workflows and execute them efficiently."+\
        " Tasks can be run locally or in a grid computing environment to increase"+\
        " efficiency. Essential information from all tasks is recorded, using the "+\
        "default logger and command line reporters, to ensure reproducibility. "+\
        "A auto-doc feature allows for workflows to generate documentation automatically"+\
        " to further ensure reproducibility by capturing the latest essential workflow "+\
        "information. AnADAMA2 was architected to be modular allowing users to "+\
        "customize the application by subclassing the base grid meta-schedulers, "+\
        "reporters, and tracked objects (ie files, executables, etc).",
    url="http://huttenhower.sph.harvard.edu/anadama2",
    keywords=['microbial','microbiome','bioinformatics','workflow','grid','anadama','anadama2'],
    platforms=['Linux','MacOS'],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
        ],
    packages=find_packages(exclude=['ez_setup', 'tests', 'tests.*']),
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'anadama2_aws_batch_task = anadama2.grid.aws_batch_task:main',
        ]
    },
    install_requires=requires,
    test_suite="tests.test_suite",
    cmdclass={ 'sphinx_build' : SphinxBuild }
)



# -*- coding: utf-8 -*-
import re
import fnmatch
import itertools

import six

from .util import matcher


class TaskContainer(list):
    """Contains tasks. Tasks can be accessed by task_no or by name"""

    def __init__(self, *args, **kwargs):
        self.by_name = dict()
        return super(TaskContainer, self).__init__(*args, **kwargs)


    def _update(self, task):
        self.by_name[task.name] = task

        
    def _get_or_search(self, key):
        if '*' in key:
            hits = list(self.search(fnmatch.translate(key)))
            if not hits:
                raise KeyError
            return hits
        return self.by_name[key]


    def search(self, q):
        return iter(val for val in self if re.search(q, val.name))


    def append(self, task):
        self._update(task)
        return super(TaskContainer, self).append(task)


    def extend(self, iterable):
        a, b = itertools.tee(iterable)
        for task in a:
            self._update(task)
        return super(TaskContainer, self).extend(b)


    def __setitem__(self, key, task):
        self._update(task)
        return super(TaskContainer, self).__setitem__(key, task)


    def __getitem__(self, key):
        try:
            if isinstance(key, six.string_types):
                return self._get_or_search(key)
            return super(TaskContainer, self).__getitem__(key)
        except KeyError:
            msg = "Unable to find task with `{}'. Perhaps you meant `{}'?"
            m = matcher.closest(key, iter(t.name for t in self))[0][1]
            raise KeyError(msg.format(key, m))
        except IndexError:
            msg = "No task with number {}. There are only {} tasks."
            raise IndexError(msg.format(key, len(self)))
 

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            if '*' in item:
                try:
                    next(self.search(fnmatch.translate(item)))
                    return True
                except StopIteration:
                    return False
            else:
                return item in self.by_name
            
        return super(TaskContainer, self).__contains__(item)



# -*- coding: utf-8 -*-
import itertools
import operator

from six.moves import zip

first = operator.itemgetter(0)

def windows(iterable, length):
    args = itertools.tee(iterable, length)
    # advance each iterator as many steps as its rank-1
    # thus, advance the 1st iterator none, 2nd iterator once. etc.
    for i in range(len(args)-1, 0, -1):
        for iter_ in args[i:]:
            next(iter_, None)
    return zip(*args)


def kmer_set(iterable, kmer_lengths=(2,)):
    ret = set()
    for k in kmer_lengths:
        ret.update(windows(iterable, k))
    return ret


def distance(a_str, b_str, kmer_lengths=(2,)):
    a_chunks = kmer_set(a_str, kmer_lengths)
    b_chunks = kmer_set(b_str, kmer_lengths)
    return sum((len(a_chunks.difference(b_chunks)),
                len(b_chunks.difference(a_chunks))))


def similarity(a_str, b_str, kmer_lengths=(2,)):
    a_chunks = kmer_set(a_str, kmer_lengths)
    b_chunks = kmer_set(b_str, kmer_lengths)
    return len(a_chunks.intersection(b_chunks))


def min_with_ties(iterable, key=None):
    iterable = iter(iterable)
    if not key:
        key = lambda val : val

    try:
        ret = [next(iterable)]
    except StopIteration:
        return ret

    for item in iterable:
        if key(item) < key(ret[0]):
            ret = [item]
        elif key(item) == key(ret[0]):
            ret.append(item)

    return ret
    

def closest(needle_str, haystack, kmer_lengths=(2,)):
    distances = [ (distance(needle_str, s, kmer_lengths), s)
                  for s in haystack ]
    return min_with_ties(distances, key=first)


def find_match(needle_str, haystack_strs, kmer_lengths=(2,)):
    haystack = set(haystack_strs)
    distances = [ (distance(needle_str, s, kmer_lengths), s)
                  for s in haystack ]
    d, match = min(distances, key=first)
    return match



# -*- coding: utf-8 -*-
import os
import re
import sys
import logging
import itertools
import time
import functools
import shlex
import copy
import threading

import six

from ..helpers import build_actions
from ..tracked import AWSHugeTrackedFile
from ..runners import _get_task_result

from .grid import Grid
from .grid import GridWorker
from .grid import GridQueue

from .aws_batch_task import parse_s3, upload_file, download_file

if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess

class AWS(Grid):
    """This class enables the Workflow class to dispatch tasks to
    AWS Batch. Use it like so:

    .. code:: python

      from anadama2 import Workflow
      from anadama2.grid import AWS

      powerup = AWS(partition="general")
      ctx = Workflow(grid=powerup)
      ctx.do("wget "
             "ftp://public-ftp.hmpdacc.org/"
             "HMMCP/finalData/hmp1.v35.hq.otu.counts.bz2 "
             "-O @{input/hmp1.v35.hq.otu.counts.bz2}")

      # run on slurm with 200 MB of memory, 4 cores, and 60 minutes
      t1 = ctx.grid_do("pbzip2 -d -p 4 < #{input/hmp1.v35.hq.otu.counts.bz2} "
                       "> @{input/hmp1.v35.hq.otu.counts}",
                       mem=200, cores=4, time=60)

      # run on slurm on the serial_requeue partition
      ctx.grid_add_task("some_huge_analysis {depends[0]} {targets[0]}",
                        depends=t1.targets, targets="output.txt",
                        mem=4000, cores=1, time=300, partition="serial_requeue")


      ctx.go()


    :param partition: The name of the SLURM partition to submit tasks to
    :type partition: str

    :param tempdir: The temp directory for scripts and job output stderr/stdout

    """

    def __init__(self, partition, tmpdir):
        super(AWS, self).__init__("aws", AWSGridWorker, AWSQueue(partition), tmpdir)

class AWSGridWorker(GridWorker):
    """ Base Grid Worker class """

    def __init__(self, work_q, result_q, lock, reporter):
        super(AWSGridWorker, self).__init__(work_q, result_q, lock, reporter)

    @classmethod
    def run_task_by_type(cls, task, extra):
        import cloudpickle
        import boto3

        (perf, tmpdir, grid_queue, reporter) = extra

        # check that task writes output to S3 bucket
        if not task.targets or not isinstance(task.targets[0], AWSHugeTrackedFile):
            result = _get_task_result(task)
            extra_error = "AWS Batch tasks must write targets to S3 buckets"
            return result._replace(error=extra_error)

        # move the temp tracked files to /tmp folder for docker run
        # update the action to use the new dependency locations        
        tracked_with_temp = list(filter(lambda x: x.temp_files(), task.depends+task.targets))
        task_working_directory = os.path.join("/tmp",task.targets[0].tmpdir)
        tmp_paths = [x.prepend_local_path(task_working_directory) for x in tracked_with_temp]

        # build the actions again with the new targets/depends paths
        task.actions=build_actions(task.actions_raw, task.depends, task.targets,
            task.visible, task.kwargs, task.use_parse_sh) 

        task_name="task_"+str(task.task_no)
        task_pkl_file_basename = task_name+".pkl"
        task_pkl_file=os.path.join(tmpdir,task_pkl_file_basename)
        task_result_pkl_file_basename = "result_"+task_name+".pkl"
        task_result_pkl_file=os.path.join(tmpdir,task_result_pkl_file_basename)

        # write the input pickle file
        resource = boto3.resource("s3")
        bucket, key, filename = parse_s3(task.targets[0])      
        input_s3 = "s3://"+bucket+"/"+task_pkl_file_basename
        output_s3 = "s3://"+bucket+"/"+task_result_pkl_file_basename

        with open(task_pkl_file,"wb") as file_handle:
            cloudpickle.dump(task,file_handle)
        # copy to cloud
        upload_file(resource,bucket,task_pkl_file_basename,task_pkl_file)

        # update the task to run the pickle script
        pickle_task = copy.deepcopy(task)
        pickle_task.actions = [" ".join(["anadama2_aws_batch_task",
            "--input", input_s3, "--output", output_s3,
            "--working-directory", task_working_directory])]

        # run the task as a command
        result = cls.run_task_command(pickle_task, extra)

        # download the results
        download_file(resource,bucket,task_result_pkl_file_basename,task_result_pkl_file)

        # decode the result
        extra_error = None
        try:
            with open(task_result_pkl_file,"rb") as file_handle:
                result = cloudpickle.load(file_handle)
        except (ValueError, EOFError, IOError):
            extra_error = "Unable to read task result, check the AWS batch logs"

        if extra_error:
            result = result._replace(error=str(result.error)+extra_error)

        return result

class AWSQueue(GridQueue):
    
    def __init__(self, partition):
        super(AWSQueue, self).__init__(partition)

        import boto3
        # get the region for this instance
        try:
            region = subprocess.check_output(["ec2-metadata","--availability-zone"]).rstrip().split(": ")[-1]
        except (subprocess.CalledProcessError, EnvironmentError):
            region = "us-east-2"

        # this is the number of seconds to wait after task definition submission
        self.submit_definition_sleep = 3

        # this is the refresh rate for checking the queue, in seconds
        self.refresh_rate = 2*60

        # this is a lock for task definition submission
        self.lock_submit_task_definition = threading.Lock()

        self.partition = partition

        self.client = boto3.client("batch",region)
       
        self.job_code_completed="SUCCEEDED"
        self.job_code_failed="FAILED"
       
        self.all_queued_codes=['SUBMITTED','PENDING','RUNNABLE','STARTING','RUNNING']
        self.all_stopped_codes=[self.job_code_completed,self.job_code_failed]
        self.all_codes=self.all_queued_codes+self.all_stopped_codes

    @staticmethod
    def job_submission_failed(jobid):
        """ Check if the job failed in submission and did not get an id """
        return True if "error" in jobid else False

    @staticmethod
    def get_job_id_from_submit_output(stdout):
        return stdout

    @staticmethod
    def submit_command(grid_script):   
        def submit_aws_batch(grid_script):
            response = grid_script()
            return response['jobId']

        return functools.partial(submit_aws_batch, grid_script)

    def create_grid_script(self,partition,cpus,minutes,memory,command,taskid,dir,docker_image):
        """ Create a grid script from the template also creating job definition """

        if not docker_image:
            docker_image='amazonlinux'

        # get the partition for the task (to allow for multiple queues and partitions in task definition)
        partition = self.get_partition(time, partition)

        # create job definition
        job_name = "task_{}".format(taskid)

        # get lock to submit job definition
        self.lock_submit_task_definition.acquire()

        response = self.client.register_job_definition(
            containerProperties={
                'image': docker_image,
                'memory': memory,
                'vcpus': cpus,
                'volumes': [
                    { 'host': { 'sourcePath': '/opt' },
                      'name': 'opt'
                    },
                    { 'host': { 'sourcePath': '/tmp' },
                      'name': 'tmp'
                    },
                ],
                'mountPoints': [
                    {
                        'containerPath': '/opt',
                        'readOnly': True,
                        'sourceVolume': 'opt'
                    },
                    {
                        'containerPath': '/tmp',
                        'readOnly': False,
                        'sourceVolume': 'tmp'
                    }
                ]
            },
            jobDefinitionName=job_name,
            type='container'
            )

        # pause after submission
        time.sleep(self.submit_definition_sleep)

        # release lock
        self.lock_submit_task_definition.release()

        submit_script = functools.partial(self.client.submit_job,
            containerOverrides={'command': shlex.split(command)},
            jobDefinition=job_name,
            jobName=job_name,
            jobQueue=partition)

        return submit_script, None, None, None
    
    def job_failed(self,status):
        # check if the job has a status that it failed
        return status == self.job_code_failed
        
    def job_stopped(self,status):
        # check if the job has a status which indicates it stopped running
        return status in self.all_stopped_codes

    def refresh_queue_status(self):
        """ Get the latest status for all the grid jobs """
   
        def gather_job_status(response):
            job_stats = []
            for job in response['jobSummaryList']:
                job_stats.append([job['jobId'],job['status'],"NA","NA","NA"])
            return job_stats
    
        def list_jobs(state, nextToken=None):
            if self.partition_long != self.partition_short:
                if nextToken:
                    response = self.client.list_jobs(jobQueue=self.partition_short,jobStatus=state,nextToken=nextToken)
                    response2 = self.client.list_jobs(jobQueue=self.partition_long,jobStatus=state,nextToken=nextToken)
                    if response2['jobSummaryList']:
                        response['jobSummaryList']+=response2['jobSummaryList']
                else:
                    response = self.client.list_jobs(jobQueue=self.partition_short,jobStatus=state)
                    response2 = self.client.list_jobs(jobQueue=self.partition_long,jobStatus=state)
                    if response2['jobSummaryList']:
                        response['jobSummaryList']+=response2['jobSummaryList']
            else:
                if nextToken:
                    response = self.client.list_jobs(jobQueue=self.partition,jobStatus=state,nextToken=nextToken)
                else:
                    response = self.client.list_jobs(jobQueue=self.partition,jobStatus=state)
            return response

        job_status = []
        for state in self.all_codes:
            response = list_jobs(state)
            job_status += gather_job_status(response)
            while response.get('nextToken',None):
                response = list_jobs(state, nextToken=response['nextToken'])
                job_status += gather_job_status(response)
        
        return job_status
    



# -*- coding: utf-8 -*-
import re
import sys
import logging
import argparse
import os
import subprocess
import collections

from .util import kebab
from .util.fname import script_wd

logger = logging.getLogger(__name__)

class Configuration(object):
    """The Configuration class makes objects that get user input via a
    command line interface and store the user input for easy access.

    Typical usage is as follows:

    .. code:: python

        from anadama2.cli import Configuration

        conf = Configuration(
            "StrainPhlan workflow"
        ).add("input", desc="Input Fastq files", type="dir",
              default="input/", pattern="*.fastq"
        ).add("output", type="dir",
              default="output/"
        ).ask_user()

        conf.input # "input/"
        conf.output # "output/"


    :keyword description: Set the description shown to the user when
      the help flag is given
    :type description: str

    :keyword version: Set the version shown to the user when the
      version flag is given
    :type version: str

    :keyword defaults: Add a set of default options to the
      Configuration object. These defaults change behavior of
      :meth:`anadama2.workflow.Workflow.go`. 
    :type defaults: bool

    """
    
    def __init__(self, description=None, version=None, defaults=True, remove_options=None, prompt_user=True):
        # set the description and version for the workflow
        self.description = description
        self.version = version
        
        # set if should prompt user for command line arguments
        self.prompt_user=prompt_user

        self._arguments = {}
        self._user_arguments = collections.OrderedDict()

        # Add the help short option so it is not selected by the function _find_short
        # when trying to find the short option for a new user option
        self._shorts = set('h')
        self._user_asked = False

        if not description:
            self.description = "AnADAMA2 Workflow"
            
        # create a parser instance
        self.parser = argparse.ArgumentParser(
            description=self.description,
            formatter_class=argparse.RawTextHelpFormatter)
        
        if defaults:
            self._arguments=collections.OrderedDict(self.get_default_options())
            for opt, info in self.get_default_options():
                self._shorts.add(info.short)

        # remove default options, if provided
        if remove_options:
            for option in remove_options:
                self.remove(option)
                
        # add a version option if version is provided
        if self.version:
            self.parser.add_argument("--version", action="version", version="%(prog)s v"+self.version)
            
            
    @staticmethod
    class Argument(object):
        def __init__(self, short, long, default=None, type=None, action=None, 
            choices=None, help=None, dest=None, required=None):
            
            # set the short and long option name
            self.short=short
            self.long=long
            
            # keep the keywords for the option that are set
            keywords={"default":default, "type":type, "action":action, "choices":choices, "help":help, "dest":dest, "required":required}
            self.keywords = {key:value for key, value in keywords.items() if value is not None}
                
    @classmethod
    def get_default_options(cls):
        return [
            ("output", cls.Argument("-o", "--output", required=True, 
                help="Write output to this directory")),
            ("input", cls.Argument("-i", "--input", default=os.getcwd(), 
                help="Find inputs in this directory \n[default: %(default)s]")),
            ("jobs", cls.Argument(None, "--local-jobs", default=1, type=int, dest="jobs", 
                help="Number of tasks to execute in parallel locally \n[default: %(default)s]")),
            ("grid_jobs", cls.Argument(None, '--grid-jobs', default=0, type=int, 
                help="Number of tasks to execute in parallel on the grid \n[default: %(default)s]")),
            ("grid", cls.Argument(None, "--grid", default=cls.identify_grid(),
                help="Run gridable tasks on this grid type \n[default: %(default)s]")),
            ("grid_partition", cls.Argument(None, "--grid-partition", default=cls.default_partitions(),
                help="Partition/queue used for gridable tasks.\nProvide a single partition or a comma-delimited list\nof short/long partitions with a cutoff.\n[default: %(default)s]")),
            ("grid_benchmark", cls.Argument(None, "--grid-benchmark", default="on" if cls.identify_grid() != "None" else "off", choices=["on","off"],
                help="Benchmark gridable tasks \n[default: %(default)s]")),
            ("grid_options", cls.Argument(None, "--grid-options", action="append",
                help="Grid specific options that will be applied to each grid task")),
            ("grid_environment", cls.Argument(None, "--grid-environment", action="append",
                help="Commands that will be run before each grid task to set up environment")),
            ("dry_run", cls.Argument(None, "--dry-run", action="store_true", 
                help="Print tasks to be run but don't execute their actions ")),
            ("skip_nothing", cls.Argument(None, "--skip-nothing", action="store_true", 
                help="Run all tasks. Rerun tasks that have already been run.")),
            ("quit_early", cls.Argument(None, '--quit-early', action="store_true",
                help="Stop if a task fails. By default,\nall tasks (except sub-tasks of failed tasks) will run.")),
            ("until_task", cls.Argument(None, "--until-task",
                help="Stop after running this task. Use task name or number.")),
            ("exclude_task", cls.Argument(None, "--exclude-task", action="append",
                help="Don't run these tasks. Add multiple times to append.")),
            ("target", cls.Argument(None, "--target", action="append",
                help="Only run tasks that generate these targets.\nAdd multiple times to append.\nPatterns with ? and * are allowed.")),
            ("exclude_target", cls.Argument(None, "--exclude-target", action="append",
                help="Don't run tasks that generate these targets.\nAdd multiple times to append.\nPatterns with ? and * are allowed.")),
            ("log_level", cls.Argument(None,"--log-level",default="INFO", choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], 
                help="Set the level of output for the log \n[default: %(default)s]"))
        ]
                
    @staticmethod
    def identify_grid():
        """ Check if a grid is found on the machine and determine which type """
        
        found_grid="None"
        # check for the grid job submission command for slurm and sge
        for command, grid in [["sbatch","slurm"],["qsub","sge"],["aws","aws"]]:
            try:
                output=subprocess.check_output(["which",command],stderr=subprocess.STDOUT)
                found_grid=grid
            except subprocess.CalledProcessError:
                pass
            
        return found_grid
    
    @classmethod
    def default_partitions(cls):
        """ Get default partitions based on the default grid """
        
        grid = cls.identify_grid()
        if grid == "slurm":
            partitions = ["serial_requeue","general",4*60]
        elif grid == "sge":
            partitions = "broad"
        elif grid == "aws":
            partitions = ["general"]
        else:
            partitions = []
            
        return ",".join(map(str,partitions))

    def add(self, name, desc=None, type=None, default=None, short=None,
            action=None, choices=None, required=None, add_short=None, add_default=None):
        """Add an option to the Configuration object.

        :param name: Set the name of the option. This name is
          transformed with :func:`anadama2.util.kebab` to the long
          option flag in the command line interface e.g. ``entry_point
          -> --entry-point``.
        :type name: str

        :keyword desc: Set the description of the option. This
          description is shown next to the option name in the help
          display shown when the user gives the ``--help`` flag.
        :type desc: str

        :keyword type: Argparse type
        :type type: str
        
        :keyword default: Set a default for the option. This value is
          stored if the user doesn't set the flag for this option.

        :keyword short: Set the short flag for this option. Defaults
          to find a short flag based off of the value for ``name``.
        :type short: str
        
        :keyword action: Set an action to execute with flag
        :type action: function or string
        
        :keyword required: Add this to the set of required options.

        :returns: self (the current Configuration object)

        """
        
        # remove special characters from the name
        name=kebab(name)
        
        # change format of final name
        option = re.sub(r'-', '_', name)
        
        # compute the short if requested
        if add_short:
            short=self._find_short(name,short)
            
        if short:
            short="-"+short
            
        # add the default to the option
        if desc and default is not None:
            desc=desc+"\n[default: %(default)s]"

        # add the new item to the end of the user arguments
        self._user_arguments[option] = self.Argument(short, "--"+name, default, type, action, 
            choices, help=desc, required=required)

        return self

    def change(self, name, **kwargs):
        """Change an option. Use keyword options to change the attribute of
        the option. 

        Here's an example:

        .. code:: python

            from anadama2 import Workflow
            from anadama2.cli import Configuration

            ctx = Workflow(vars=Configuration().change("output", default="."))
            ...

        """
        
        for key, value in kwargs.items():
            self._arguments[name][key]=value

    def remove(self, name):
        """ Remove an option from the Configuration object.
        :param name: The name of the option to remove
        
        """
        
        # remove the short if set
        try:
            self._shorts.remove(self._arguments[name].short)
        except KeyError:
            pass
        
        del self._arguments[name]


    def get(self, name, default=None):
        """Get a stored option value from the Configuration object.
        
        :param name: The name of the value to get
        """
        
        # get arguments from the command line (will not run again if already parsed)
        if not self._user_asked:
            self.ask_user()

        return getattr(self, name, default)
    
    def get_option_values(self):
        """Get all of the values for the command line options. Prompt user if not asked."""
        
        class CommandLineOptions(object):
            def __getattr__(self, name):
                # if an attribute can not be found, this is the last function called
                all_option_names=", ".join(vars(self).keys())
                error_message="Unable to find option '{0}' in command line options.\n".format(name)
                error_message+="The available options are: {0}".format(all_option_names)
                raise AttributeError(error_message)
            
        # get arguments from the command line (will not run again if already parsed)
        if not self._user_asked:
            self.ask_user()
        
        args=CommandLineOptions()
        for option in list(self._user_arguments.keys()) + list(self._arguments.keys()):
            option = re.sub(r'-', '_', option)
            value = self.get(option)
            setattr(args,option,value)
                
        return args


    __getitem__ = get


    def reset(self):
        """Invalidate caching behavior of
        :meth:`anadama2.cli.Configuration.ask_user`. Use this method if
        you want to re-ask the user for input.

        """
        self._user_asked = False


    def ask_user(self, override=False, argv=None):
        """Set up and display the command line interface. Store defaults and
        any user input in the Configuration object. The defaults and
        user input are cached; if this method is called again, it does
        nothing and returns the current Configuration object.

        :keyword override: Override the caching behavior.
        :type override: bool

        :keyword argv: The command line arguments to parse. Defaults
          to sys.argv[1:].
        :type argv: list of str

        :returns: self (the current Configuration object)

        """
        # if not set to prompt user, then use defaults
        if not self.prompt_user:
            for arg_name, arg_values in self._arguments.items():
                setattr(self,arg_name,arg_values.keywords.get("default",None))
            return self
        
        if self._user_asked and not override:
            return self
        
        # add options by dictionary order starting with the user arguments
        for arg_name, arg_values in list(self._user_arguments.items()) + list(self._arguments.items()):
            if arg_values.short:
                self.parser.add_argument(arg_values.short, arg_values.long, **arg_values.keywords)
            else:
                self.parser.add_argument(arg_values.long, **arg_values.keywords)
            
        opts = self.parser.parse_args(args=argv)
        for name in list(self._user_arguments.keys()) + list(self._arguments.keys()):
            name = re.sub(r'-', '_', name)
            val = getattr(opts, name)
            logger.info("Command line argument `%s' = `%s'", name, val)
            setattr(self, name, val)
        self._user_asked = True
        return self

        
    def _find_short(self, name, short=None):
        s = None
        if short and short[-1] not in self._shorts:
            s = short[-1]
        for char in name.lower() + name.upper():
            char_ascii_value = ord(char)
            if 64 < char_ascii_value < 91 or 96 < char_ascii_value < 123 and char not in self._shorts:
                s = char
                break
        if s is None:
            raise ValueError("Unable to find short option flag for "+name)
        self._shorts.add(s)
        return "-"+s

    



#!/usr/bin/env python

""" A script to run a pickled task placed in a AWS S3 bucket """

import os
import sys
import argparse
import shutil

import six

from anadama2.runners import _run_task_locally
from anadama2.tracked import try_get_local_path

def parse_arguments(args):
    """
    Parse the arguments from the user
    """
    parser = argparse.ArgumentParser(
        description= "AnADAMA2 AWS Batch task runner\n",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "--input",
        required=True,
        help="the pickle file of task information\n")
    parser.add_argument(
        "--output",
        required=True,
        help="the pickle file of task results\n")
    parser.add_argument(
        "--working-directory",
        default="/tmp",
        help="the directory to run the task from\n")

    return parser.parse_args()

def parse_s3(name):
    """ Parse the file into bucket, key, and filename """

    bucket = str(name).replace("s3://","").split("/")[0]
    key = "/".join(str(name).replace("s3://","").split("/")[1:])
    filename = key.split("/")[-1]

    return bucket, key, filename

def download_file(resource,bucket,key,filename):
    """ Download a file from s3 """
    import botocore

    create_directories(filename)

    try:
        resource.Bucket(bucket).download_file(key,filename)
    except botocore.exceptions.ClientError:
        print("Unable to download file: s://"+bucket+"/"+key)

def upload_file(resource,bucket,key,filename):
    """ Upload a file to s3 """

    resource.Bucket(bucket).upload_file(filename,key)

def local_path(filename,working_directory):
    return filename.replace("s3://",working_directory)

def create_directories(filename):
    """ Create directories if needed """

    directory = os.path.dirname(filename)
    if directory and not os.path.isdir(directory):
        os.makedirs(directory)

def main():
    import cloudpickle
    import boto3
    resource = boto3.resource("s3")

    # Parse arguments from the command line
    args=parse_arguments(sys.argv)

    # move to the working directory
    os.makedirs(args.working_directory)
    os.chdir(args.working_directory)

    # download the input file from S3
    in_bucket, in_key, in_filename = parse_s3(args.input)
    local_in_file = local_path(in_filename, args.working_directory)
    download_file(resource,in_bucket,in_key,local_in_file)
    
    # read in the task information
    task = cloudpickle.load(open(local_in_file,"rb"))

    # create folders for targets (if needed)
    for target in task.targets:
        target_local_path = try_get_local_path(target)
        if isinstance(target_local_path, basestring):
            create_directories(target_local_path)

    # run the task
    result = _run_task_locally(task)

    # write the pickled results to a file
    out_bucket, out_key, out_filename = parse_s3(args.output)
    local_out_file = local_path(out_filename, args.working_directory)
    cloudpickle.dump(result,open(local_out_file,"wb"))

    # upload the results to S3
    upload_file(resource,out_bucket,out_key,local_out_file)

    # remove the working directory
    shutil.rmtree(args.working_directory)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
class AuthInfo(object):
    def __init__(self, username, projectname, commit_id, auth_key):
        self.username    = username
        self.projectname = projectname
        self.commit_id   = commit_id
        self.auth_key    = auth_key


    @classmethod    
    def parse(cls, s):
        if not s:
            return None
        try:
            return cls(*s.split("_:_"))
        except:
            return None


    def to_dict(self):
        return {
            "username": self.username, "projectname": self.projectname,
            "commit_id": self.commit_id, "auth_key": self.auth_key
        }
